import logging

from regipy.hive_types import SYSTEM_HIVE_TYPE
from regipy.plugins.plugin import Plugin
from regipy.exceptions import RegistryKeyNotFoundException


logger = logging.getLogger(__name__)


PROCESSOR_PATH = r"Control\Session Manager\Environment"
processor_list = (
    "PROCESSOR_ARCHITECTURE",
    "NUMBER_OF_PROCESSORS",
    "PROCESSOR_IDENTIFIER",
    "PROCESSOR_REVISION",
)


class ProcessorArchitecturePlugin(Plugin):
    NAME = "processor_architecture"
    DESCRIPTION = "Get processor architecture info from the System's environment key"
    COMPATIBLE_HIVE = SYSTEM_HIVE_TYPE

    def can_run(self):
        return self.registry_hive.hive_type == SYSTEM_HIVE_TYPE

    def run(self):
        self.entries = {}
        processor_subkeys = self.registry_hive.get_control_sets(PROCESSOR_PATH)
        for processor_subkey in processor_subkeys:
            try:
                processor = self.registry_hive.get_key(processor_subkey)
            except RegistryKeyNotFoundException as ex:
                logger.error(
                    f"Could not find {self.NAME} subkey at {processor_subkey}: {ex}"
                )
                continue
            self.entries[processor_subkey] = {}
            for val in processor.iter_values():
                if val.name in processor_list:
                    self.entries[processor_subkey][val.name] = val.value



# flake8: noqa
from .amcache.amcache import AmCachePlugin
from .bcd.boot_entry_list import BootEntryListPlugin
from .ntuser.installed_programs_ntuser import InstalledProgramsNTUserPlugin
from .ntuser.network_drives import NetworkDrivesPlugin
from .ntuser.persistence import NTUserPersistencePlugin
from .ntuser.shellbags_ntuser import ShellBagNtuserPlugin
from .ntuser.tsclient import TSClientPlugin
from .ntuser.typed_urls import TypedUrlsPlugin
from .ntuser.typed_paths import TypedPathsPlugin
from .ntuser.user_assist import UserAssistPlugin
from .ntuser.winrar import WinRARPlugin
from .ntuser.winscp_saved_sessions import WinSCPSavedSessionsPlugin
from .ntuser.word_wheel_query import WordWheelQueryPlugin
from .sam.local_sid import LocalSidPlugin
from .security.domain_sid import DomainSidPlugin
from .software.classes_installer import SoftwareClassesInstallerPlugin
from .software.image_file_execution_options import ImageFileExecutionOptions
from .software.installed_programs import InstalledProgramsSoftwarePlugin
from .software.last_logon import LastLogonPlugin
from .software.persistence import SoftwarePersistencePlugin
from .software.printdemon import PrintDemonPlugin
from .software.profilelist import ProfileListPlugin
from .software.tracing import RASTracingPlugin
from .software.uac import UACStatusPlugin
from .system.active_controlset import ActiveControlSetPlugin
from .system.bam import BAMPlugin
from .system.bootkey import BootKeyPlugin
from .system.computer_name import ComputerNamePlugin
from .system.host_domain_name import HostDomainNamePlugin
from .system.routes import RoutesPlugin
from .system.safeboot_configuration import SafeBootConfigurationPlugin
from .system.services import ServicesPlugin
from .system.shimcache import ShimCachePlugin
from .system.timezone_data import TimezoneDataPlugin
from .system.usbstor import USBSTORPlugin
from .system.wdigest import WDIGESTPlugin
from .usrclass.shellbags_usrclass import ShellBagUsrclassPlugin
from .ntuser.classes_installer import NtuserClassesInstallerPlugin
from .system.network_data import NetworkDataPlugin
from .software.winver import WinVersionPlugin
from .system.previous_winver import PreviousWinVersionPlugin
from .system.shutdown import ShutdownPlugin
from .system.processor_architecture import ProcessorArchitecturePlugin
from .system.crash_dump import CrashDumpPlugin
from .software.susclient import SusclientPlugin
from .system.disablelastaccess import DisableLastAccessPlugin
from .system.codepage import CodepagePlugin
from .software.disablesr import DisableSRPlugin
from .system.diag_sr import DiagSRPlugin
from .software.spp_clients import SppClientsPlugin
from .system.backuprestore import BackupRestorePlugin
from .system.timezone_data2 import TimezoneDataPlugin2
from .ntuser.wsl import WSLPlugin


import binascii
import logging
import datetime as dt

import pytz
from click import progressbar
from typing import Iterator

from regipy import RegistryHive, NKRecord, Subkey
from regipy.utils import MAX_LEN


logger = logging.getLogger(__name__)


def get_filtered_subkeys(
    registry_hive: RegistryHive,
    name_key_entry: NKRecord,
    start_date: str = None,
    end_date: str = None,
    verbose=False,
    fetch_values=True,
) -> Iterator[NKRecord]:
    """
    Get records filtered by the specified timestamps
    :param registry_hive: A RegistryHive object
    :param name_key_entry: A list of paths as strings
    :param start_date: Include only subkeys modified after the specified date
                     in isoformat UTC, for example: 2020-02-18T14:15:00.000000
    :param end_date: Include only subkeys modified before the specified date
                     in isoformat UTC, for example: 2020-02-20T14:15:00.000000
    """
    skipped_entries_count = 0
    if start_date:
        start_date = pytz.utc.localize(dt.datetime.fromisoformat(start_date))

    if end_date:
        end_date = pytz.utc.localize(dt.datetime.fromisoformat(end_date))

    with progressbar(
        registry_hive.recurse_subkeys(name_key_entry, fetch_values=False)
    ) as reg_subkeys:
        for subkey_count, subkey in enumerate(reg_subkeys):
            if start_date:
                if subkey.timestamp < start_date:
                    skipped_entries_count += 1
                    logger.debug(
                        f"Skipping entry {subkey} which has a timestamp prior to start_date"
                    )
                    continue

            if end_date:
                if subkey.timestamp > end_date:
                    skipped_entries_count += 1
                    logger.debug(
                        f"Skipping entry {subkey} which has a timestamp after the end_date"
                    )
                    continue

            nk = registry_hive.get_key(subkey.path)
            yield Subkey(
                subkey_name=subkey.subkey_name,
                path=subkey.path,
                timestamp=subkey.timestamp,
                values=list(nk.iter_values(as_json=True)) if fetch_values else [],
                values_count=subkey.values_count,
            )
        logger.info(
            f"{skipped_entries_count} out of {subkey_count} subkeys were filtered out due to timestamp constrains"
        )


def _normalize_subkey_fields(field) -> str:
    if isinstance(field, bytes):
        return binascii.b2a_hex(field[:MAX_LEN])
    elif isinstance(field, dt.datetime):
        return field.isoformat()
    return field



import struct
import binascii

import logging

from regipy.hive_types import SYSTEM_HIVE_TYPE
from regipy.plugins.plugin import Plugin
from regipy.utils import convert_wintime
from regipy.exceptions import RegistryKeyNotFoundException


logger = logging.getLogger(__name__)

TZ_DATA_PATH = r"Control\TimeZoneInformation"


class TimezoneDataPlugin2(Plugin):
    NAME = "timezone_data2"
    DESCRIPTION = "Get timezone data"
    COMPATIBLE_HIVE = SYSTEM_HIVE_TYPE

    def run(self):
        self.entries = {}
        tzdata_subkeys = self.registry_hive.get_control_sets(TZ_DATA_PATH)
        for tzdata_subkey in tzdata_subkeys:
            try:
                tzdata = self.registry_hive.get_key(tzdata_subkey)
            except RegistryKeyNotFoundException as ex:
                logger.error(
                    f"Could not find {self.NAME} subkey at {tzdata_subkey}: {ex}"
                )
                continue
            self.entries[tzdata_subkey] = [
                x for x in tzdata.iter_values(as_json=self.as_json)
            ]
            self.entries[tzdata_subkey] = {
                "last_write": convert_wintime(tzdata.header.last_modified).isoformat()
            }
            for val in tzdata.iter_values():
                if val.name in ("ActiveTimeBias", "Bias", "DaylightBias"):
                    self.entries[tzdata_subkey][val.name] = struct.unpack(
                        ">l", struct.pack(">L", val.value & 0xFFFFFFFF)
                    )[0]
                else:
                    self.entries[tzdata_subkey][val.name] = val.value



import logging

from regipy.hive_types import SYSTEM_HIVE_TYPE
from regipy.plugins.plugin import Plugin
from regipy.utils import convert_wintime

logger = logging.getLogger(__name__)

HOST_PARAMETERS_PATH = r"Services\Tcpip\Parameters"


class HostDomainNamePlugin(Plugin):
    NAME = "host_domain_name"
    DESCRIPTION = "Get the computer host and domain names"
    COMPATIBLE_HIVE = SYSTEM_HIVE_TYPE

    def run(self):
        logger.debug("Started Host and Domain Name Plugin...")

        for subkey_path in self.registry_hive.get_control_sets(HOST_PARAMETERS_PATH):
            subkey = self.registry_hive.get_key(subkey_path)

            hostname = subkey.get_value("Hostname", as_json=self.as_json)
            domain = subkey.get_value("Domain", as_json=self.as_json)

            # The default key value is 0x00000000 (REG_DWORD) when
            # the Windows machine is not in an AD domain.
            if not isinstance(domain, str):
                domain = None

            self.entries.append(
                {
                    "hostname": hostname,
                    "domain": domain,
                    "timestamp": convert_wintime(
                        subkey.header.last_modified, as_json=self.as_json
                    ),
                }
            )



import logging

from inflection import underscore

from regipy import RegistryKeyNotFoundException, convert_wintime
from regipy.hive_types import SOFTWARE_HIVE_TYPE
from regipy.plugins.plugin import Plugin

logger = logging.getLogger(__name__)

LAST_LOGON_KEY_PATH = r"\Microsoft\Windows\CurrentVersion\Authentication\LogonUI"


class LastLogonPlugin(Plugin):
    NAME = "last_logon_plugin"
    DESCRIPTION = "Get the last logged on username"
    COMPATIBLE_HIVE = SOFTWARE_HIVE_TYPE

    def run(self):
        try:
            subkey = self.registry_hive.get_key(LAST_LOGON_KEY_PATH)
        except RegistryKeyNotFoundException as ex:
            logger.error(
                f"Could not find {self.NAME} subkey at {LAST_LOGON_KEY_PATH}: {ex}"
            )
            return None

        self.entries = {
            "last_write": convert_wintime(
                subkey.header.last_modified, as_json=self.as_json
            ),
            **{
                underscore(x.name): x.value
                for x in subkey.iter_values(as_json=self.as_json)
            },
        }



import logging

from regipy.hive_types import SYSTEM_HIVE_TYPE
from regipy.plugins.plugin import Plugin
from regipy.utils import convert_wintime
from regipy.exceptions import RegistryKeyNotFoundException


logger = logging.getLogger(__name__)


PROCESSOR_PATH = r"Control\Nls\CodePage"
crash_items = "ACP"


class CodepagePlugin(Plugin):
    NAME = "codepage"
    DESCRIPTION = "Get codepage value"
    COMPATIBLE_HIVE = SYSTEM_HIVE_TYPE

    def can_run(self):
        return self.registry_hive.hive_type == SYSTEM_HIVE_TYPE

    def run(self):
        self.entries = {}
        codepage_subkeys = self.registry_hive.get_control_sets(PROCESSOR_PATH)
        for codepage_subkey in codepage_subkeys:
            try:
                codepage = self.registry_hive.get_key(codepage_subkey)
            except RegistryKeyNotFoundException as ex:
                logger.error(
                    f"Could not find {self.NAME} subkey at {codepage_subkey}: {ex}"
                )
                continue
            self.entries[codepage_subkey] = {
                "last_write": convert_wintime(codepage.header.last_modified).isoformat()
            }
            for val in codepage.iter_values():
                if val.name in crash_items:
                    self.entries[codepage_subkey][val.name] = val.value



import logging

from regipy.hive_types import NTUSER_HIVE_TYPE
from regipy.plugins.plugin import Plugin

logger = logging.getLogger(__name__)


class TemplatePlugin(Plugin):
    NAME = "template_plugin"
    DESCRIPTION = "template_description"

    def can_run(self):
        # TODO: Choose the relevant condition - to determine if the plugin is relevant for the given hive
        return self.registry_hive.hive_type == NTUSER_HIVE_TYPE

    def run(self):
        # TODO: Return the relevant values
        raise NotImplementedError



import logging

from regipy.hive_types import NTUSER_HIVE_TYPE
from regipy.plugins.plugin import Plugin
from regipy.utils import get_subkey_values_from_list

logger = logging.getLogger(__name__)

PERSISTENCE_ENTRIES = [
    r"\Software\Microsoft\Windows NT\CurrentVersion\Run",
    r"\Software\Microsoft\Windows NT\CurrentVersion\Terminal Server\Install\Software\Microsoft\Windows\CurrentVersion\Run",
    r"\Software\Microsoft\Windows NT\CurrentVersion\Terminal Server\Install\Software\Microsoft\Windows\CurrentVersion\RunOnce",
    r"\Software\Microsoft\Windows\CurrentVersion\Policies\Explorer\Run",
    r"\Software\Microsoft\Windows\CurrentVersion\Run",
    r"\Software\Microsoft\Windows\CurrentVersion\RunOnce",
    r"\Software\Microsoft\Windows\CurrentVersion\RunOnceEx",
    r"\Software\Microsoft\Windows\CurrentVersion\RunOnce\Setup",
    r"\Software\Microsoft\Windows\CurrentVersion\RunServices",
    r"\Software\Microsoft\Windows\CurrentVersion\RunServicesOnce",
    r"\Software\Wow6432Node\Microsoft\Windows\CurrentVersion\Policies\Explorer\Run",
    r"\Software\Wow6432Node\Microsoft\Windows\CurrentVersion\Run",
    r"\Software\Wow6432Node\Microsoft\Windows\CurrentVersion\RunOnce",
    r"\Software\Wow6432Node\Microsoft\Windows\CurrentVersion\RunOnceEx",
    r"\Software\Wow6432Node\Microsoft\Windows\CurrentVersion\RunOnce\Setup",
    r"\Software\Microsoft\Windows NT\CurrentVersion\Winlogon\Notify",
]


class NTUserPersistencePlugin(Plugin):
    NAME = "ntuser_persistence"
    DESCRIPTION = "Retrieve values from known persistence subkeys in NTUSER hive"
    COMPATIBLE_HIVE = NTUSER_HIVE_TYPE

    def run(self):
        self.entries = get_subkey_values_from_list(
            self.registry_hive,
            PERSISTENCE_ENTRIES,
            as_json=self.as_json,
            trim_values=False,
        )



import logging
from datetime import datetime, timezone

from regipy.exceptions import RegistryKeyNotFoundException
from regipy.hive_types import SYSTEM_HIVE_TYPE
from regipy.plugins.plugin import Plugin
from regipy.utils import convert_wintime

logger = logging.getLogger(__name__)

INTERFACES_PATH = r"Services\Tcpip\Parameters\Interfaces"


class NetworkDataPlugin(Plugin):
    NAME = "network_data"
    DESCRIPTION = "Get network data from many interfaces"
    COMPATIBLE_HIVE = SYSTEM_HIVE_TYPE

    def get_network_info(self, subkey, interfaces=None):
        if interfaces is None:
            interfaces = []

        try:
            for interface in subkey.iter_subkeys():
                entries = {
                    "interface_name": interface.name,
                    "last_modified": convert_wintime(
                        interface.header.last_modified, as_json=self.as_json
                    ),
                    "incomplete_data": False,  # New key to indicate incomplete data
                    "dhcp_enabled": interface.get_value("EnableDHCP")
                    == 1,  # Boolean value
                }

                if entries["dhcp_enabled"]:
                    entries.update(
                        {
                            "dhcp_server": interface.get_value("DhcpServer"),
                            "dhcp_ip_address": interface.get_value("DhcpIPAddress"),
                            "dhcp_subnet_mask": interface.get_value("DhcpSubnetMask"),
                            "dhcp_default_gateway": interface.get_value(
                                "DhcpDefaultGateway"
                            ),
                            "dhcp_name_server": interface.get_value("DhcpNameServer"),
                            "dhcp_domain": interface.get_value("DhcpDomain"),
                        }
                    )

                    # Lease Obtained Time
                    lease_obtained_time = interface.get_value("LeaseObtainedTime")
                    if lease_obtained_time is not None:
                        try:
                            lease_obtained_time_str = datetime.fromtimestamp(
                                lease_obtained_time, timezone.utc
                            ).strftime("%Y-%m-%d %H:%M:%S")
                            entries["dhcp_lease_obtained_time"] = (
                                lease_obtained_time_str
                            )
                        except (OSError, ValueError) as e:
                            logger.error(
                                f"Error converting DHCP lease obtained time for interface {interface.name}: {e}"
                            )
                            entries["incomplete_data"] = True

                    # Lease Terminates Time
                    lease_terminates_time = interface.get_value("LeaseTerminatesTime")
                    if lease_terminates_time is not None:
                        try:
                            lease_terminates_time_str = datetime.fromtimestamp(
                                lease_terminates_time, timezone.utc
                            ).strftime("%Y-%m-%d %H:%M:%S")
                            entries["dhcp_lease_terminates_time"] = (
                                lease_terminates_time_str
                            )
                        except (OSError, ValueError) as e:
                            logger.error(
                                f"Error converting DHCP lease terminates time for interface {interface.name}: {e}"
                            )
                            entries["incomplete_data"] = True

                else:
                    entries.update(
                        {
                            "ip_address": interface.get_value("IPAddress"),
                            "subnet_mask": interface.get_value("SubnetMask"),
                            "default_gateway": interface.get_value("DefaultGateway"),
                            "name_server": interface.get_value("NameServer"),
                            "domain": interface.get_value("Domain"),
                        }
                    )

                try:
                    if interface.subkey_count:
                        sub_interfaces = []
                        sub_interfaces = self.get_network_info(
                            interface, sub_interfaces
                        )
                        entries["sub_interface"] = sub_interfaces
                except Exception as e:
                    logger.error(
                        f"Error processing sub-interfaces for interface {interface.name}: {e}"
                    )
                    entries["incomplete_data"] = True

                interfaces.append(entries)

        except Exception as e:
            logger.error(f"Error iterating over subkeys in {subkey.path}: {e}")

        return interfaces

    def run(self):
        self.entries = {}

        for control_set_interfaces_path in self.registry_hive.get_control_sets(
            INTERFACES_PATH
        ):
            try:
                subkey = self.registry_hive.get_key(control_set_interfaces_path)
            except RegistryKeyNotFoundException as ex:
                logger.error(
                    f"Registry key not found at path {control_set_interfaces_path}: {ex}"
                )
                continue  # Skip to the next path if the key is not found

            try:
                self.entries[control_set_interfaces_path] = {
                    "timestamp": convert_wintime(
                        subkey.header.last_modified, as_json=self.as_json
                    )
                }
                interfaces = []
                interfaces = self.get_network_info(subkey, interfaces)
                self.entries[control_set_interfaces_path]["interfaces"] = interfaces
            except Exception as ex:
                logger.error(
                    f"Error processing registry key {control_set_interfaces_path}: {ex}"
                )
                self.entries[control_set_interfaces_path]["incomplete_data"] = True



import csv
import json
import logging
import os
import time


import attr
import click
from tabulate import tabulate

from regipy.plugins.plugin import PLUGINS
from regipy.recovery import apply_transaction_logs
from regipy.regdiff import compare_hives
from regipy.plugins.utils import run_relevant_plugins
from regipy.registry import RegistryHive
from regipy.exceptions import RegistryKeyNotFoundException
from regipy.utils import calculate_xor32_checksum, _setup_logging
from regipy.cli_utils import get_filtered_subkeys, _normalize_subkey_fields

logger = logging.getLogger(__name__)


@click.command()
@click.argument(
    "hive_path",
    type=click.Path(exists=True, dir_okay=False, resolve_path=True),
    required=True,
)
@click.option("-v", "--verbose", is_flag=True, default=True, help="Verbosity")
def parse_header(hive_path, verbose):
    _setup_logging(verbose=verbose)
    registry_hive = RegistryHive(hive_path)

    click.secho(tabulate(registry_hive.header.items(), tablefmt="fancy_grid"))

    if (
        registry_hive.header.primary_sequence_num
        != registry_hive.header.secondary_sequence_num
    ):
        click.secho("Hive is not clean! You should apply transaction logs", fg="red")

    calculated_checksum = calculate_xor32_checksum(registry_hive._stream.read(508))
    if registry_hive.header.checksum != calculated_checksum:
        click.secho("Hive is not clean! Header checksum does not match", fg="red")


@click.command()
@click.argument(
    "hive_path",
    type=click.Path(exists=True, dir_okay=False, resolve_path=True),
    required=True,
)
@click.option(
    "-o",
    "output_path",
    type=click.Path(exists=False, dir_okay=False, resolve_path=True),
    required=False,
)
@click.option("-p", "--registry-path", help="A registry path to start iterating from")
@click.option(
    "-t",
    "--timeline",
    is_flag=True,
    default=False,
    help="Create a CSV timeline instead",
)
@click.option(
    "-l",
    "--hive-type",
    type=click.STRING,
    required=False,
    help="Specify a hive type, if it could not be identified for some reason",
)
@click.option(
    "-r",
    "--partial_hive_path",
    type=click.STRING,
    required=False,
    help='The path from which the partial hive actually starts, for example: -t ntuser -r "/Software" '
    "would mean this is actually a HKCU hive, starting from HKCU/Software",
)
@click.option("-v", "--verbose", is_flag=True, default=False, help="Verbosity")
@click.option(
    "-d",
    "--do-not-fetch-values",
    is_flag=True,
    default=False,
    help="Not fetching the values for each subkey "
    "makes the iteration way faster. Values count will still be returned",
)
@click.option(
    "-s",
    "--start-date",
    type=click.STRING,
    required=False,
    help='If "-s" was specified, fetch only values for subkeys starting this timestamp in isoformat',
)
@click.option(
    "-e",
    "--end-date",
    type=click.STRING,
    required=False,
    help='If "-e" was specified, fetch only values for subkeys until this timestamp in isoformat',
)
def registry_dump(
    hive_path,
    output_path,
    registry_path,
    timeline,
    hive_type,
    partial_hive_path,
    verbose,
    do_not_fetch_values,
    start_date,
    end_date,
):
    _setup_logging(verbose=verbose)
    registry_hive = RegistryHive(
        hive_path, hive_type=hive_type, partial_hive_path=partial_hive_path
    )

    start_time = time.monotonic()

    if registry_path:
        try:
            name_key_entry = registry_hive.get_key(registry_path)
        except RegistryKeyNotFoundException as ex:
            logger.debug("Did not find the key: {}".format(ex))
            return
    else:
        name_key_entry = registry_hive.root

    if timeline and not output_path:
        click.secho(
            "You must provide an output path if choosing timeline output!", fg="red"
        )
        return

    if output_path:
        with open(output_path, "w") as output_file:
            if timeline:
                csvwriter = csv.DictWriter(
                    output_file,
                    delimiter=",",
                    quotechar='"',
                    quoting=csv.QUOTE_MINIMAL,
                    fieldnames=["timestamp", "subkey_name", "values_count", "values"],
                )
                csvwriter.writeheader()

            for subkey_count, entry in enumerate(
                get_filtered_subkeys(
                    registry_hive,
                    name_key_entry,
                    fetch_values=not do_not_fetch_values,
                    start_date=start_date,
                    end_date=end_date,
                )
            ):
                if timeline:
                    csvwriter.writerow(
                        {
                            "subkey_name": entry.path,
                            "timestamp": entry.timestamp,
                            "values_count": entry.values_count,
                            "values": entry.values,
                        }
                    )
                else:
                    output_file.write(
                        json.dumps(
                            attr.asdict(
                                entry,
                            ),
                            separators=(
                                ",",
                                ":",
                            ),
                            default=_normalize_subkey_fields,
                        )
                    )
                    output_file.write("\n")
    else:
        for subkey_count, entry in enumerate(
            registry_hive.recurse_subkeys(
                name_key_entry, as_json=True, fetch_values=not do_not_fetch_values
            )
        ):
            click.secho(json.dumps(attr.asdict(entry), indent=4))

    click.secho(
        f"Completed in {time.monotonic() - start_time}s ({subkey_count} subkeys enumerated)"
    )


@click.command()
@click.argument(
    "hive_path",
    type=click.Path(exists=True, dir_okay=False, resolve_path=True),
    required=True,
)
@click.option(
    "-o",
    "output_path",
    type=click.Path(exists=False, dir_okay=False, resolve_path=True),
    required=True,
    help="Output path for plugins result",
)
@click.option(
    "-p",
    "--plugins",
    type=click.STRING,
    required=False,
    help="A plugin or list of plugins to execute command separated",
)
@click.option(
    "-t",
    "--hive-type",
    type=click.STRING,
    required=False,
    help="Specify a hive type, if it could not be identified for some reason",
)
@click.option(
    "-r",
    "--partial_hive_path",
    type=click.STRING,
    required=False,
    help='The path from which the partial hive actually starts, for example: -t ntuser -r "/Software" '
    "would mean this is actually a HKCU hive, starting from HKCU/Software",
)
@click.option("-v", "--verbose", is_flag=True, default=False, help="Verbosity")
def run_plugins(hive_path, output_path, plugins, hive_type, partial_hive_path, verbose):
    _setup_logging(verbose=verbose)
    registry_hive = RegistryHive(
        hive_path, hive_type=hive_type, partial_hive_path=partial_hive_path
    )
    click.secho("Loaded {} plugins".format(len(PLUGINS)), fg="white")

    if plugins:
        plugin_names = {x.NAME for x in PLUGINS}
        plugins = plugins.split(",")
        plugins = set(plugins)
        if not plugins.issubset(plugin_names):
            click.secho(
                "Invalid plugin names given: {}".format(
                    ",".join(set(plugins) - plugin_names)
                ),
                fg="red",
            )
            click.secho(
                "Use --help or -h to get list of plugins and their descriptions",
                fg="red",
            )
            return

    # Run relevant plugins
    plugin_results = run_relevant_plugins(registry_hive, as_json=True, plugins=plugins)

    # If output path was set, dump results to disk
    if output_path:
        with open(output_path, "w") as f:
            f.write(json.dumps(plugin_results, indent=4))
    else:
        print(json.dumps(plugin_results, indent=4))
    click.secho(
        "Finished: {}/{} plugins matched the hive type".format(
            len(plugin_results), len(PLUGINS)
        ),
        fg="green",
    )


@click.command()
def list_plugins():
    click.secho(
        tabulate(
            [(x.NAME, x.COMPATIBLE_HIVE, x.DESCRIPTION) for x in PLUGINS],
            headers=["Plugin Name", "Compatible hive", "Description"],
        )
    )


@click.command()
@click.argument(
    "first_hive_path",
    type=click.Path(exists=True, dir_okay=False, resolve_path=True),
    required=True,
)
@click.argument(
    "second_hive_path",
    type=click.Path(exists=True, dir_okay=False, resolve_path=True),
    required=True,
)
@click.option(
    "-o",
    "output_path",
    type=click.Path(exists=False, dir_okay=False, resolve_path=True),
    required=False,
)
@click.option("-v", "--verbose", is_flag=True, default=False, help="Verbosity")
def reg_diff(first_hive_path, second_hive_path, output_path, verbose):
    _setup_logging(verbose=verbose)
    REGDIFF_HEADERS = ["difference", "first_hive", "second_hive", "description"]

    found_differences = compare_hives(
        first_hive_path, second_hive_path, verbose=verbose
    )
    click.secho(
        "Comparing {} vs {}".format(
            os.path.basename(first_hive_path), os.path.basename(second_hive_path)
        )
    )

    if output_path:
        with open(output_path, "w") as csvfile:
            csvwriter = csv.writer(csvfile, delimiter="|", quoting=csv.QUOTE_MINIMAL)
            csvwriter.writerow(REGDIFF_HEADERS)
            for difference in found_differences:
                csvwriter.writerow(difference)
    else:
        click.secho(
            tabulate(found_differences, headers=REGDIFF_HEADERS, tablefmt="fancy_grid")
        )
    click.secho(f"Detected {len(found_differences)} differences", fg="green")


@click.command()
@click.argument(
    "hive_path",
    type=click.Path(exists=True, dir_okay=False, resolve_path=True),
    required=True,
)
@click.option(
    "-p",
    "primary_log_path",
    type=click.Path(exists=True, dir_okay=False, resolve_path=True),
    required=True,
)
@click.option(
    "-s",
    "secondary_log_path",
    type=click.Path(exists=True, dir_okay=False, resolve_path=True),
    required=False,
)
@click.option(
    "-o",
    "output_path",
    type=click.Path(exists=False, dir_okay=False, resolve_path=True),
    required=False,
)
@click.option("-v", "--verbose", is_flag=True, default=True, help="Verbosity")
def parse_transaction_log(
    hive_path, primary_log_path, secondary_log_path, output_path, verbose
):
    _setup_logging(verbose=verbose)
    logger.info(f"Processing hive {hive_path} with transaction log {primary_log_path}")
    if secondary_log_path:
        logger.info(
            f"Processing hive {hive_path} with secondary transaction log {secondary_log_path}"
        )

    restored_hive_path, recovered_dirty_pages_count = apply_transaction_logs(
        hive_path,
        primary_log_path,
        secondary_log_path=secondary_log_path,
        restored_hive_path=output_path,
        verbose=verbose,
    )
    if recovered_dirty_pages_count:
        click.secho(
            f"Recovered {recovered_dirty_pages_count} dirty pages. Restored hive is at {restored_hive_path}",
            fg="green",
        )



import logging
from datetime import datetime, timezone

from regipy.hive_types import SOFTWARE_HIVE_TYPE
from regipy.plugins.plugin import Plugin
from regipy.utils import convert_wintime
from regipy.exceptions import RegistryKeyNotFoundException

logger = logging.getLogger(__name__)


WIN_VER_PATH = r"\Microsoft\Windows NT\CurrentVersion"
os_list = (
    "ProductName",
    "ReleaseID",
    "CSDVersion",
    "CurrentVersion",
    "CurrentBuild",
    "CurrentBuildNumber",
    "InstallationType",
    "EditionID",
    "ProductName",
    "ProductId",
    "BuildLab",
    "BuildLabEx",
    "CompositionEditionID",
    "RegisteredOrganization",
    "RegisteredOwner",
    "InstallDate",
)


class WinVersionPlugin(Plugin):
    NAME = "winver_plugin"
    DESCRIPTION = "Get relevant OS information"
    COMPATIBLE_HIVE = SOFTWARE_HIVE_TYPE

    def can_run(self):
        return self.registry_hive.hive_type == SOFTWARE_HIVE_TYPE

    def run(self):
        logger.info("Started winver Plugin...")

        try:
            key = self.registry_hive.get_key(WIN_VER_PATH)
        except RegistryKeyNotFoundException as ex:
            logger.error(f"Could not find {self.NAME} subkey at {WIN_VER_PATH}: {ex}")
            return None

        self.entries = {
            WIN_VER_PATH: {
                "last_write": convert_wintime(key.header.last_modified).isoformat()
            }
        }

        for val in key.iter_values():
            if val.name in os_list:
                if val.name == "InstallDate":
                    self.entries[WIN_VER_PATH][val.name] = datetime.fromtimestamp(
                        val.value, timezone.utc
                    ).strftime("%Y-%m-%d %H:%M:%S")
                else:
                    self.entries[WIN_VER_PATH][val.name] = val.value



import logging

from regipy import RegistryKeyNotFoundException
from regipy.hive_types import NTUSER_HIVE_TYPE
from regipy.plugins.plugin import Plugin
from regipy.utils import convert_wintime

logger = logging.getLogger(__name__)

WINSCP_SAVED_SESSIONS_PATH = r"\Software\Martin Prikryl\WinSCP 2\Sessions"


class WinSCPSavedSessionsPlugin(Plugin):
    NAME = "winscp_saved_sessions"
    DESCRIPTION = "Retrieve list of WinSCP saved sessions"
    COMPATIBLE_HIVE = NTUSER_HIVE_TYPE

    def _get_winscp_saved_sessions(self, subkey_path):
        try:
            sessions_sk = self.registry_hive.get_key(subkey_path)
        except RegistryKeyNotFoundException as ex:
            logger.error(ex)
            return

        for winscp_saved_session in sessions_sk.iter_subkeys():
            values = (
                {
                    x.name: x.value
                    for x in winscp_saved_session.iter_values(as_json=self.as_json)
                }
                if winscp_saved_session.values_count
                else {}
            )
            self.entries.append(
                {
                    "timestamp": convert_wintime(
                        winscp_saved_session.header.last_modified, as_json=self.as_json
                    ),
                    "hive_name": "HKEY_CURRENT_USER",
                    "key_path": rf"HKEY_CURRENT_USER{subkey_path}\{winscp_saved_session.name}",
                    **values,
                }
            )

    def run(self):
        self._get_winscp_saved_sessions(WINSCP_SAVED_SESSIONS_PATH)



import logging

from construct import CString, GreedyRange, Int32ul

from regipy.exceptions import RegistryKeyNotFoundException
from regipy.hive_types import NTUSER_HIVE_TYPE
from regipy.plugins.plugin import Plugin
from regipy.utils import convert_wintime

logger = logging.getLogger(__name__)

WORD_WHEEL_QUERY_KEY_PATH = (
    r"\Software\Microsoft\Windows\CurrentVersion\Explorer\WordWheelQuery"
)


class WordWheelQueryPlugin(Plugin):
    NAME = "word_wheel_query"
    DESCRIPTION = "Parse the word wheel query artifact"
    COMPATIBLE_HIVE = NTUSER_HIVE_TYPE

    def run(self):
        try:
            subkey = self.registry_hive.get_key(WORD_WHEEL_QUERY_KEY_PATH)
        except RegistryKeyNotFoundException as ex:
            logger.error(
                f"Could not find {self.NAME} plugin data at: {WORD_WHEEL_QUERY_KEY_PATH}: {ex}"
            )
            return None

        timestamp = convert_wintime(subkey.header.last_modified, as_json=self.as_json)

        mru_list_order = subkey.get_value("MRUListEx")

        # If this is the value, the list is empty
        if mru_list_order == 0xFFFFFFFF:
            return None

        for i, entry_name in enumerate(GreedyRange(Int32ul).parse(mru_list_order)):
            entry_value = subkey.get_value(str(entry_name))

            if not entry_value:
                continue

            self.entries.append(
                {
                    "last_write": timestamp,
                    "mru_id": entry_name,
                    "order": i,
                    "name": CString("utf-16").parse(entry_value),
                }
            )



import logging
import attr

from regipy.hive_types import SYSTEM_HIVE_TYPE
from regipy.plugins.plugin import Plugin


logger = logging.getLogger(__name__)

SELECT = r"\Select"


class ActiveControlSetPlugin(Plugin):
    NAME = "active_control_set"
    DESCRIPTION = "Get information on SYSTEM hive control sets"
    COMPATIBLE_HIVE = SYSTEM_HIVE_TYPE

    def run(self):
        subkey = self.registry_hive.get_key(SELECT)
        self.entries = [x for x in subkey.iter_values(as_json=self.as_json)]
        if self.as_json:
            self.entries = [attr.asdict(x) for x in self.entries]



# flake8: noqa
import cProfile
import io
import lzma
import os
import pstats
from contextlib import contextmanager
from pathlib import Path
from pstats import SortKey
from tempfile import mktemp

import logging

from regipy.registry import RegistryHive

logger = logging.getLogger(__name__)


@contextmanager
def profiling():
    pr = cProfile.Profile()
    pr.enable()
    yield
    pr.disable()
    s = io.StringIO()
    sortby = SortKey.CUMULATIVE
    ps = pstats.Stats(pr, stream=s).sort_stats(sortby)
    ps.print_stats()
    print(s.getvalue())


@contextmanager
def get_file_from_tests(file_name):
    path = str(Path(__file__).parent.joinpath("data").joinpath(file_name))
    tempfile_path = mktemp()
    with open(tempfile_path, "wb") as tmp:
        with lzma.open(path) as f:
            tmp.write(f.read())
    yield tempfile_path
    os.remove(tempfile_path)


registry_path = "SYSTEM_2.xz"
print(f"Iterating over all subkeys in {registry_path}")
with profiling():
    with get_file_from_tests(registry_path) as reg:
        registry_hive = RegistryHive(reg)
        keys = [x for x in registry_hive.recurse_subkeys(fetch_values=False)]
print(f"Done.")

"""

=== Fetching subkey values:

Iterating over all subkeys in SYSTEM_2.xz
         6681282 function calls (6548808 primitive calls) in 21.406 seconds

   Ordered by: cumulative time

   ncalls  tottime  percall  cumtime  percall filename:lineno(function)
        1    0.009    0.009   21.440   21.440 /Users/martin/Projects/regipy/regipy_tests/profiling.py:47(<listcomp>)
115672/20016    0.326    0.000   21.431    0.001 /Users/martin/Projects/regipy/regipy/registry.py:126(recurse_subkeys)
    66844    1.067    0.000   19.524    0.000 /Users/martin/Projects/regipy/regipy/registry.py:413(iter_values)
    49779    0.280    0.000   10.845    0.000 /Users/martin/Projects/regipy/regipy/registry.py:378(read_value)
  1101844   10.603    0.000   10.603    0.000 {method 'read' of '_io.BytesIO' objects}
    30464    0.357    0.000    5.337    0.000 /Users/martin/Projects/regipy/regipy/utils.py:149(try_decode_binary)
   116294    3.587    0.000    5.021    0.000 {method 'decode' of 'bytes' objects}
   150968    0.252    0.000    2.785    0.000 /opt/anaconda3/envs/regipy/lib/python3.9/site-packages/construct/core.py:290(parse_stream)
187671/150968    0.117    0.000    2.359    0.000 /opt/anaconda3/envs/regipy/lib/python3.9/site-packages/construct/core.py:311(_parsereport)
    74151    0.051    0.000    1.779    0.000 /opt/anaconda3/envs/regipy/lib/python3.9/site-packages/construct/core.py:786(_parse)
    42243    0.024    0.000    1.434    0.000 /opt/anaconda3/envs/regipy/lib/python3.9/encodings/utf_16_le.py:15(decode)
    24363    0.039    0.000    1.413    0.000 /Users/martin/Projects/regipy/regipy/registry.py:322(iter_subkeys)
    42243    1.410    0.000    1.410    0.000 {built-in method _codecs.utf_16_le_decode}
    24367    0.116    0.000    1.333    0.000 /Users/martin/Projects/regipy/regipy/registry.py:350(_parse_subkeys)
    49779    0.026    0.000    0.943    0.000 :79(parseall)
    49779    0.579    0.000    0.917    0.000 :26(parse_struct_1)
    20015    0.127    0.000    0.837    0.000 /Users/martin/Projects/regipy/regipy/registry.py:292(__init__)
    20015    0.012    0.000    0.584    0.000 :124(parseall)
    20015    0.378    0.000    0.572    0.000 :23(parse_struct_1)
    

=== Without fetching subkeys:

Iterating over all subkeys in SYSTEM_2.xz
         2068855 function calls (1973084 primitive calls) in 5.288 seconds

   Ordered by: cumulative time

   ncalls  tottime  percall  cumtime  percall filename:lineno(function)
        1    0.037    0.037    4.969    4.969 /Users/martin/Projects/regipy/regipy_tests/profiling.py:47(<listcomp>)
115672/20016    0.462    0.000    4.932    0.000 /Users/martin/Projects/regipy/regipy/registry.py:126(recurse_subkeys)
    24363    0.097    0.000    4.104    0.000 /Users/martin/Projects/regipy/regipy/registry.py:324(iter_subkeys)
    24367    0.312    0.000    3.900    0.000 /Users/martin/Projects/regipy/regipy/registry.py:352(_parse_subkeys)
    48737    0.322    0.000    3.121    0.000 /opt/anaconda3/envs/regipy/lib/python3.9/site-packages/construct/core.py:290(parse_stream)
    48737    0.119    0.000    2.640    0.000 /opt/anaconda3/envs/regipy/lib/python3.9/site-packages/construct/core.py:311(_parsereport)
    20015    0.343    0.000    2.370    0.000 /Users/martin/Projects/regipy/regipy/registry.py:294(__init__)
    24372    0.038    0.000    2.334    0.000 /opt/anaconda3/envs/regipy/lib/python3.9/site-packages/construct/core.py:786(_parse)
    20015    0.033    0.000    1.683    0.000 :124(parseall)
    20015    1.123    0.000    1.650    0.000 :23(parse_struct_1)
     4353    0.045    0.000    0.610    0.000 :76(parseall)

"""



import logging

from regipy.hive_types import SYSTEM_HIVE_TYPE
from regipy.plugins.plugin import Plugin
from regipy.utils import convert_wintime
from regipy.exceptions import RegistryKeyNotFoundException

logger = logging.getLogger(__name__)


PROCESSOR_PATH = r"Control\CrashControl"
crash_items = ("CrashDumpEnabled", "DumpFile", "MinidumpDir", "LogEvent")
dump_enabled = {
    "0": "None",
    "1": "Complete memory dump",
    "2": "Kernel memory dump",
    "3": "Small memory dump (64 KB)",
    "7": "Automatic memory dump",
}


class CrashDumpPlugin(Plugin):
    NAME = "crash_dump"
    DESCRIPTION = "Get crash control information"
    COMPATIBLE_HIVE = SYSTEM_HIVE_TYPE

    def can_run(self):
        return self.registry_hive.hive_type == SYSTEM_HIVE_TYPE

    def run(self):
        self.entries = {}
        architecture_subkeys = self.registry_hive.get_control_sets(PROCESSOR_PATH)
        for architecture_subkey in architecture_subkeys:
            try:
                architecture = self.registry_hive.get_key(architecture_subkey)
            except RegistryKeyNotFoundException as ex:
                logger.error(
                    f"Could not find {self.NAME} subkey at {architecture_subkey}: {ex}"
                )
                continue
            self.entries[architecture_subkey] = {
                "last_write": convert_wintime(
                    architecture.header.last_modified
                ).isoformat()
            }
            for val in architecture.iter_values():
                if val.name in crash_items:
                    self.entries[architecture_subkey][val.name] = val.value
                if val.name == "CrashDumpEnabled":
                    self.entries[architecture_subkey]["CrashDumpEnabledStr"] = (
                        dump_enabled.get(str(val.value), "")
                    )



import logging

from regipy import RegistryKeyNotFoundException, convert_wintime
from regipy.hive_types import SOFTWARE_HIVE_TYPE
from regipy.plugins.plugin import Plugin

logger = logging.getLogger(__name__)

PORTS_KEY_PATH = r"\Microsoft\Windows NT\CurrentVersion\Ports"


class PrintDemonPlugin(Plugin):
    NAME = "print_demon_plugin"
    DESCRIPTION = "Get list of installed printer ports, as could be taken advantage by cve-2020-1048"
    COMPATIBLE_HIVE = SOFTWARE_HIVE_TYPE

    def run(self):
        try:
            subkey = self.registry_hive.get_key(PORTS_KEY_PATH)
        except RegistryKeyNotFoundException as ex:
            logger.error(f"Could not find {self.NAME} subkey at {PORTS_KEY_PATH}: {ex}")
            return None

        last_write = convert_wintime(subkey.header.last_modified).isoformat()
        for port in subkey.iter_values():
            self.entries.append(
                {
                    "timestamp": last_write,
                    "port_name": port.name,
                    "parameters": (
                        port.value
                        if isinstance(port.value, int)
                        else port.value.split(",")
                    ),
                }
            )



import logging

from regipy.hive_types import SOFTWARE_HIVE_TYPE
from regipy.plugins.plugin import Plugin
from regipy.utils import get_subkey_values_from_list

logger = logging.getLogger(__name__)

PERSISTENCE_ENTRIES = [
    r"\Microsoft\Windows\CurrentVersion\Policies\Explorer\Run",
    r"\Microsoft\Windows\CurrentVersion\Run",
    r"\Microsoft\Windows\CurrentVersion\RunOnce",
    r"\Microsoft\Windows\CurrentVersion\RunOnce\Setup",
    r"\Microsoft\Windows\CurrentVersion\RunOnceEx",
    r"\Wow6432Node\Microsoft\Windows\CurrentVersion\Run",
    r"\Wow6432Node\Microsoft\Windows\CurrentVersion\RunOnce",
    r"\Wow6432Node\Microsoft\Windows\CurrentVersion\RunOnce\Setup",
    r"\Wow6432Node\Microsoft\Windows\CurrentVersion\RunOnceEx",
    r"\Wow6432Node\Microsoft\Windows\CurrentVersion\Policies\Explorer\Run",
    r"\Software\Microsoft\Windows NT\CurrentVersion\Winlogon\Notify",
]


class SoftwarePersistencePlugin(Plugin):
    NAME = "software_plugin"
    DESCRIPTION = "Retrieve values from known persistence subkeys in Software hive"
    COMPATIBLE_HIVE = SOFTWARE_HIVE_TYPE

    def run(self):
        self.entries = get_subkey_values_from_list(
            self.registry_hive, PERSISTENCE_ENTRIES, as_json=self.as_json
        )



"""
The boot key is an encryption key that is stored in
the Windows SYSTEM registry hive.

This key is used by several Windows components to encrypt
sensitive information like the AD database,
machine account password or system certificates etc.
"""

import logging

from regipy.registry import NKRecord
from regipy.hive_types import SYSTEM_HIVE_TYPE
from regipy.plugins.plugin import Plugin
from regipy.utils import convert_wintime

logger = logging.getLogger(__name__)

LSA_KEY_PATH = r"Control\Lsa"


def _collect_bootkey(lsa_key: NKRecord) -> str:
    """
    Extracts the 128-bit scrambled boot key from four "secret"
    LSA subkeys as a 32-character hex string.
    """

    # The boot key is taken from four separate keys:
    # SYSTEM\CurrentControlSet\Control\Lsa\{JD,Skew1,GBG,Data}.
    # However, the actual data needed is stored in a hidden field of the key
    # that cannot be seen using tools like regedit.
    # Specifically, each part of the key is stored in the key's Class attribute,
    # and is stored as a Unicode string giving the hex value of that piece of the key.

    bootkey_subkeys = {
        "JD": "",
        "Skew1": "",
        "GBG": "",
        "Data": "",
    }

    for subkey in lsa_key.iter_subkeys():
        if subkey.name in bootkey_subkeys:
            bootkey_subkeys[subkey.name] = subkey.get_class_name()

    return "".join(bootkey_subkeys.values())


# Permutation matrix for the boot key
# fmt: off
_BOOTKEY_PBOX = [
    0x8, 0x5, 0x4, 0x2,
    0xB, 0x9, 0xD, 0x3,
    0x0, 0x6, 0x1, 0xC,
    0xE, 0xA, 0xF, 0x7
]
# fmt: on


def _descramble_bootkey(key: str) -> bytes:
    """
    Parses the 128-bit binary boot key from a hex string
    and reverses the bytewise scrambling transform.
    """

    binkey = bytes.fromhex(key)

    return bytes([binkey[i] for i in _BOOTKEY_PBOX])


class BootKeyPlugin(Plugin):
    """
    The boot key is an encryption key that is stored in
    the Windows SYSTEM registry hive.
    """

    NAME = "bootkey"
    DESCRIPTION = "Get the Windows boot key"
    COMPATIBLE_HIVE = SYSTEM_HIVE_TYPE

    def run(self):
        logger.debug("Started BootKey Plugin...")

        for subkey_path in self.registry_hive.get_control_sets(LSA_KEY_PATH):
            lsa_key = self.registry_hive.get_key(subkey_path)

            bootkey = _descramble_bootkey(_collect_bootkey(lsa_key))

            self.entries.append(
                {
                    "key": bootkey.hex() if self.as_json else bootkey,
                    "timestamp": convert_wintime(
                        lsa_key.header.last_modified, as_json=self.as_json
                    ),
                }
            )



import logging

from regipy.exceptions import RegistryKeyNotFoundException
from regipy.hive_types import SOFTWARE_HIVE_TYPE
from regipy.plugins.plugin import Plugin
from regipy.utils import convert_wintime, convert_filetime


logger = logging.getLogger(__name__)

PROFILE_LIST_KEY_PATH = r"\Microsoft\Windows NT\CurrentVersion\ProfileList"


class ProfileListPlugin(Plugin):
    NAME = "profilelist_plugin"
    DESCRIPTION = "Parses information about user profiles found in the ProfileList key"
    COMPATIBLE_HIVE = SOFTWARE_HIVE_TYPE

    def run(self):
        logger.debug("Started profile list plugin...")
        try:
            subkey = self.registry_hive.get_key(PROFILE_LIST_KEY_PATH)
        except RegistryKeyNotFoundException as ex:
            logger.error(ex)

        for profile in subkey.iter_subkeys():
            self.entries.append(
                {
                    "last_write": convert_wintime(
                        profile.header.last_modified, as_json=self.as_json
                    ),
                    "path": profile.get_value("ProfileImagePath"),
                    "flags": profile.get_value("Flags"),
                    "full_profile": profile.get_value("FullProfile"),
                    "state": profile.get_value("State"),
                    "sid": profile.name,
                    "load_time": convert_filetime(
                        profile.get_value("ProfileLoadTimeLow"),
                        profile.get_value("ProfileLoadTimeHigh"),
                    ),
                    "local_load_time": convert_filetime(
                        profile.get_value("LocalProfileLoadTimeLow"),
                        profile.get_value("LocalProfileLoadTimeHigh"),
                    ),
                }
            )



import logging

from regipy.hive_types import SYSTEM_HIVE_TYPE
from regipy.plugins.plugin import Plugin
from regipy.utils import convert_wintime
from regipy.exceptions import RegistryKeyNotFoundException


logger = logging.getLogger(__name__)


BACKUPRESTORE_PATH = [
    r"Control\BackupRestore\FilesNotToSnapshot",
    r"Control\BackupRestore\FilesNotToBackup",
    r"Control\BackupRestore\KeysNotToRestore",
]


class BackupRestorePlugin(Plugin):
    NAME = "backuprestore_plugin"
    DESCRIPTION = "Gets the contents of the FilesNotToSnapshot, KeysNotToRestore, and FilesNotToBackup keys"
    COMPATIBLE_HIVE = SYSTEM_HIVE_TYPE

    def can_run(self):
        return self.registry_hive.hive_type == SYSTEM_HIVE_TYPE

    def run(self):
        self.entries = {}
        for br_path in BACKUPRESTORE_PATH:
            br_subkeys = self.registry_hive.get_control_sets(br_path)
            for br_subkey in br_subkeys:
                try:
                    backuprestore = self.registry_hive.get_key(br_subkey)
                except RegistryKeyNotFoundException as ex:
                    logger.error(
                        f"Could not find {self.NAME} subkey at {br_subkey}: {ex}"
                    )
                    continue
                self.entries[br_subkey] = {
                    "last_write": convert_wintime(
                        backuprestore.header.last_modified
                    ).isoformat()
                }
                for val in backuprestore.iter_values():
                    self.entries[br_subkey][val.name] = val.value



import logging
from regipy.exceptions import RegistryKeyNotFoundException
from regipy.hive_types import NTUSER_HIVE_TYPE
from regipy.plugins.plugin import Plugin
from regipy.utils import convert_wintime
from regipy.constants import KNOWN_GUIDS
import re

logger = logging.getLogger(__name__)

NTUSER_SHELLBAG = "\\Software\\Microsoft\\Windows\\Shell\\BagMRU"
DEFAULT_CODEPAGE = "cp1252"


class ShellBagNtuserPlugin(Plugin):
    NAME = "ntuser_shellbag_plugin"
    DESCRIPTION = "Parse NTUSER Shellbag items"
    COMPATIBLE_HIVE = NTUSER_HIVE_TYPE

    @staticmethod
    def _parse_mru(mru_val):
        mru_order_string = ""
        if isinstance(mru_val, bytes):
            mru_val = mru_val[:-4]
            for i in range(0, len(mru_val), 4):
                current_val = int.from_bytes(mru_val[i: i + 4], byteorder="little")
                mru_order_string += f"{current_val}-"

            return mru_order_string[:-1]
        else:
            return mru_order_string

    @staticmethod
    def _get_shell_item_type(shell_item):
        try:
            import pyfwsi
        except ModuleNotFoundError as ex:
            logger.exception(
                "Plugin `shellbag_plugin` has missing modules, install regipy using"
                " `pip install regipy[full]` in order to install plugin dependencies. "
                "This might take some time... "
            )
            raise ex

        if isinstance(shell_item, pyfwsi.volume):
            item_type = "Volume"

        elif isinstance(shell_item, pyfwsi.file_entry):
            item_type = "Directory"

        elif isinstance(shell_item, pyfwsi.network_location):
            item_type = "Network Location"

        elif isinstance(shell_item, pyfwsi.root_folder):
            item_type = "Root Folder"

        elif isinstance(shell_item, pyfwsi.control_panel_category):
            item_type = "Control Panel Category"

        elif isinstance(shell_item, pyfwsi.control_panel_item):
            item_type = "Control Panel Item"

        elif isinstance(shell_item, pyfwsi.users_property_view):
            item_type = "Users Property View"

        else:
            item_type = "unknown"

        return item_type

    @staticmethod
    def _check_known_guids(guid):
        if guid in KNOWN_GUIDS:
            path_segment = KNOWN_GUIDS[guid]
        else:
            path_segment = "{{{0:s}}}".format(guid)
        return path_segment

    @staticmethod
    def _get_entry_string(fwps_record):
        if fwps_record.entry_name:
            entry_string = fwps_record.entry_name
        else:
            entry_string = f"{fwps_record.entry_type:d}"
        return entry_string

    @staticmethod
    def _parse_shell_item_path_segment(self, shell_item):
        """Parses a shell item path segment.
        Args:
          shell_item (pyfwsi.item): shell item.
        Returns:
          str: shell item path segment.
        """

        try:
            import pyfwsi
            import pyfwps
        except ModuleNotFoundError as ex:
            logger.exception(
                f"Plugin `shellbag_plugin` has missing modules, install regipy using"
                f" `pip install regipy[full]` in order to install plugin dependencies. "
                f"This might take some time... "
            )
            raise ex

        path_segment = None
        full_path = None
        location_description = None

        if isinstance(shell_item, pyfwsi.volume):
            if shell_item.name:
                path_segment = shell_item.name
            elif shell_item.identifier:
                path_segment = self._check_known_guids(shell_item.identifier)

        elif isinstance(shell_item, pyfwsi.file_entry):
            long_name = ""
            for extension_block in shell_item.extension_blocks:
                if isinstance(extension_block, pyfwsi.file_entry_extension):
                    long_name = extension_block.long_name

            if long_name:
                path_segment = long_name
            elif shell_item.name:
                path_segment = shell_item.name

        elif isinstance(shell_item, pyfwsi.network_location):
            if shell_item.location:
                path_segment = shell_item.location
            if shell_item.description:
                location_description = shell_item.description
                if shell_item.comments:
                    location_description += f", {shell_item.comments}"

        elif isinstance(shell_item, pyfwsi.root_folder):
            if shell_item.shell_folder_identifier in KNOWN_GUIDS:
                path_segment = KNOWN_GUIDS[shell_item.shell_folder_identifier]
            elif (
                hasattr(shell_item, "identifier")
                and shell_item.identifier in KNOWN_GUIDS
            ):
                path_segment = KNOWN_GUIDS[shell_item.identifier]
            else:
                path_segment = "{{{0:s}}}".format(shell_item.shell_folder_identifier)

        elif isinstance(shell_item, pyfwsi.users_property_view):
            # Users property view
            if shell_item.delegate_folder_identifier in KNOWN_GUIDS:
                path_segment = KNOWN_GUIDS[shell_item.delegate_folder_identifier]
            elif (
                hasattr(shell_item, "identifier")
                and shell_item.identifier in KNOWN_GUIDS
            ):
                path_segment = KNOWN_GUIDS[shell_item.identifier]

            # Variable: Users property view
            elif shell_item.property_store_data:
                fwps_store = pyfwps.store()
                fwps_store.copy_from_byte_stream(shell_item.property_store_data)

                for fwps_set in iter(fwps_store.sets):
                    if fwps_set.identifier == "b725f130-47ef-101a-a5f1-02608c9eebac":
                        for fwps_record in iter(fwps_set.records):
                            entry_string = self._get_entry_string(fwps_record)

                            # PKEY_DisplayName: {b725f130-47ef-101a-a5f1-02608c9eebac}/10
                            if entry_string == "10":
                                if fwps_record.value_type == 0x0001:
                                    value_string = "<VT_NULL>"
                                elif fwps_record.value_type in (
                                    0x0003,
                                    0x0013,
                                    0x0014,
                                    0x0015,
                                ):
                                    value_string = str(
                                        fwps_record.get_data_as_integer()
                                    )
                                elif fwps_record.value_type in (0x0008, 0x001E, 0x001F):
                                    value_string = fwps_record.get_data_as_string()
                                elif fwps_record.value_type == 0x000B:
                                    value_string = str(
                                        fwps_record.get_data_as_boolean()
                                    )
                                elif fwps_record.value_type == 0x0040:
                                    filetime = fwps_record.get_data_as_integer()
                                    value_string = self._FormatFiletimeValue(filetime)
                                elif fwps_record.value_type == 0x0042:
                                    # TODO: add support
                                    value_string = "<VT_STREAM>"
                                elif fwps_record.value_type == 0x0048:
                                    value_string = fwps_record.get_data_as_guid()
                                elif fwps_record.value_type & 0xF000 == 0x1000:
                                    # TODO: add support
                                    value_string = "<VT_VECTOR>"
                                else:
                                    value_string = None

                                path_segment = value_string

                    elif fwps_set.identifier == "28636aa6-953d-11d2-b5d6-00c04fd918d0":
                        for fwps_record in iter(fwps_set.records):
                            entry_string = self._get_entry_string(fwps_record)

                            # PKEY_ParsingPath: {28636aa6-953d-11d2-b5d6-00c04fd918d0}/30
                            if entry_string == "30":
                                full_path = fwps_record.get_data_as_string()

        elif isinstance(shell_item, pyfwsi.control_panel_category):
            path_segment = self._check_known_guids(str(shell_item.identifier))

        elif isinstance(shell_item, pyfwsi.control_panel_item):
            path_segment = self._check_known_guids(shell_item.identifier)

        if path_segment is None:
            path_segment = "<UNKNOWN: 0x{0:02x}>".format(shell_item.class_type)

        return path_segment, full_path, location_description

    def iter_sk(self, key, reg_path, codepage=DEFAULT_CODEPAGE, base_path="", path=""):
        try:
            import pyfwsi
        except ModuleNotFoundError as ex:
            logger.exception(
                f"Plugin `shellbag_plugin` has missing modules, install regipy using"
                f" `pip install regipy[full]` in order to install plugin dependencies. "
                f"This might take some time... "
            )
            raise ex

        last_write = convert_wintime(key.header.last_modified, as_json=True)

        mru_val = key.get_value("MRUListEx")
        mru_order = self._parse_mru(mru_val)
        base_path = path

        if key.get_value("NodeSlot"):
            node_slot = str(key.get_value("NodeSlot"))
        else:
            node_slot = ""

        for v in key.iter_values(trim_values=False):
            if re.match(r"\d+", v.name):
                slot = v.name
                byte_stream = v.value
                shell_items = pyfwsi.item_list()
                shell_items.copy_from_byte_stream(byte_stream, ascii_codepage=codepage)
                for item in shell_items.items:
                    shell_type = self._get_shell_item_type(item)
                    value, full_path, location_description = (
                        self._parse_shell_item_path_segment(self, item)
                    )
                    if not path:
                        path = value
                        base_path = ""
                    else:
                        path += f"\\{value}"

                    creation_time = None
                    access_time = None
                    modification_time = None

                    if len(item.extension_blocks) > 0:
                        for extension_block in item.extension_blocks:
                            if isinstance(extension_block, pyfwsi.file_entry_extension):
                                try:
                                    creation_time = extension_block.get_creation_time()
                                    if self.as_json:
                                        creation_time = creation_time.isoformat()
                                except OSError:
                                    logger.exception(
                                        f"Malformed creation time for {path}"
                                    )
                                try:
                                    access_time = extension_block.get_access_time()
                                    if self.as_json:
                                        access_time = access_time.isoformat()
                                except OSError:
                                    logger.exception(
                                        f"Malformed access time for {path}"
                                    )

                    try:
                        if hasattr(item, "modification_time"):
                            modification_time = item.get_modification_time()
                            if self.as_json:
                                modification_time = modification_time.isoformat()
                    except OSError:
                        logger.exception(f"Malformed modification time for {path}")

                    value_name = v.name
                    mru_order_location = mru_order.split("-").index(value_name)
                    entry = {
                        "value": value,
                        "slot": slot,
                        "reg_path": reg_path,
                        "value_name": value_name,
                        "node_slot": node_slot,
                        "shell_type": shell_type,
                        "path": path,
                        "full path": full_path if full_path else None,
                        "location description": (
                            location_description if location_description else None
                        ),
                        "creation_time": creation_time,
                        "access_time": access_time,
                        "modification_time": modification_time,
                        "last_write": last_write,
                        "mru_order": mru_order,
                        "mru_order_location": mru_order_location,
                    }

                    self.entries.append(entry)
                    sk_reg_path = f"{reg_path}\\{value_name}"
                    sk = self.registry_hive.get_key(sk_reg_path)
                    self.iter_sk(sk, sk_reg_path, codepage, base_path, path)
                    path = base_path

    def run(self, codepage=DEFAULT_CODEPAGE):

        try:
            # flake8: noqa
            import pyfwsi
        except ModuleNotFoundError as ex:
            logger.exception(
                "Plugin `shellbag_plugin` has missing modules, install regipy using"
                " `pip install regipy[full]` in order to install plugin dependencies. "
                "This might take some time... "
            )
            raise ex

        try:
            shellbag_ntuser_subkey = self.registry_hive.get_key(NTUSER_SHELLBAG)
            self.iter_sk(shellbag_ntuser_subkey, NTUSER_SHELLBAG, codepage=codepage)
        except RegistryKeyNotFoundException as ex:
            logger.error(
                f"Could not find {self.NAME} plugin data at: {NTUSER_SHELLBAG}: {ex}"
            )



from construct import (
    Array,
    Bytes,
    Const,
    Enum,
    FlagsEnum,
    Int8ul,
    Int16ul,
    Int32ul,
    Int64ul,
    PaddedString,
    Struct,
    this,
)

REGF_HEADER = Struct(
    "signature" / Const(b"regf"),
    "primary_sequence_num" / Int32ul,
    "secondary_sequence_num" / Int32ul,
    "last_modification_time" / Int64ul,
    "major_version" / Int32ul,
    "minor_version" / Int32ul,
    "file_type" / Int32ul,
    "file_format" / Int32ul,
    "root_key_offset" / Int32ul,
    "hive_bins_data_size" / Int32ul,
    "clustering_factor" / Int32ul,
    "file_name" / PaddedString(64, "utf-16-le"),
    "padding" * Bytes(396),
    "checksum" / Int32ul,
).compile()

REGF_HEADER_SIZE = 4096

HBIN_HEADER = Struct(
    "signature" / Const(b"hbin"),
    "offset" / Int32ul,
    "size" / Int32ul,
    "unknown" * Int32ul,
    "unknown" * Int32ul,
    "timestamp" / Int64ul,
    "unknown" * Int32ul,
).compile()

CM_KEY_NODE_SIZE = 76
CM_KEY_NODE = Struct(
    "flags"
    / FlagsEnum(
        Int16ul,
        KEY_VOLATILE=0x0001,
        KEY_HIVE_EXIT=0x0002,
        KEY_HIVE_ENTRY=0x0004,
        KEY_NO_DELETE=0x0008,
        KEY_SYM_LINK=0x0010,
        KEY_COMP_NAME=0x0020,
        KEY_PREDEF_HANDLE=0x0040,
    ),
    "last_modified" / Int64ul,
    "access_bits" / Bytes(4),
    "parent_key_offset" / Int32ul,
    "subkey_count" / Int32ul,
    "volatile_subkey_count" / Int32ul,
    "subkeys_list_offset" / Int32ul,
    "volatile_subkeys_list_offset" / Int32ul,
    "values_count" / Int32ul,
    "values_list_offset" / Int32ul,
    "security_key_offset" / Int32ul,
    "class_name_offset" / Int32ul,
    "largest_sk_name" / Int32ul,
    "largest_sk_class_name" / Int32ul,
    "largest_value_name" / Int32ul,
    "largest_value_data" / Int32ul,
    "unknown" * Bytes(4),
    "key_name_size" / Int16ul,
    "class_name_size" / Int16ul,
    "key_name_string" / Bytes(this.key_name_size),
).compile()

SUBKEY_LIST_HEADER = Struct(
    "signature" / Bytes(2),
    "element_count" / Int16ul,
).compile()

HASH_LEAF = Struct(
    "element_count" / Int16ul,
    "elements"
    / Array(
        this.element_count,
        Struct(
            "key_node_offset" / Int32ul,
            "name_hash" / Int32ul,
        ),
    ),
).compile()

FAST_LEAF = Struct(
    "element_count" / Int16ul,
    "elements"
    / Array(
        this.element_count,
        Struct(
            "key_node_offset" / Int32ul,
            "name_hint" / Int32ul,
        ),
    ),
).compile()

LEAF_INDEX_SIGNATURE = b"li"
INDEX_LEAF = Struct(
    "element_count" / Int16ul,
    "elements"
    / Array(
        this.element_count,
        Struct(
            "key_node_offset" / Int32ul,
        ),
    ),
).compile()

INDEX_ROOT_SIGNATURE = b"ri"
INDEX_ROOT = Struct(
    "element_count" / Int16ul,
    "elements"
    / Array(
        this.element_count,
        Struct(
            "subkey_list_offset" / Int32ul,
        ),
    ),
).compile()

KEY_SECURITY = Struct(
    "reserved" * Bytes(2),
    "forward_link" / Int32ul,
    "static_link" / Int32ul,
    "reference_count" / Int32ul,
    "security_descriptor_size" / Int32ul,
    "security_descriptor" / Bytes(this.security_descriptor_size),
).compile()

VALUE_TYPE_ENUM = Enum(
    Int32ul,
    REG_NONE=0,
    REG_SZ=1,
    REG_EXPAND_SZ=2,
    REG_BINARY=3,
    REG_DWORD=4,
    REG_DWORD_BIG_ENDIAN=5,
    REG_LINK=6,
    REG_MULTI_SZ=7,
    REG_RESOURCE_LIST=8,
    REG_FULL_RESOURCE_DESCRIPTOR=9,
    REG_RESOURCE_REQUIREMENTS_LIST=10,
    REG_QWORD=11,
    REG_FILETIME=16,
)

VALUE_KEY = Struct(
    "signature" / Const(b"vk"),
    "name_size" / Int16ul,
    "data_size" / Int32ul,
    "data_offset" / Int32ul,
    "data_type" / VALUE_TYPE_ENUM,
    "flags" / FlagsEnum(Int16ul, VALUE_COMP_NAME=0x0001),
    "padding" * Int16ul,
    "name" / Bytes(this.name_size),
).compile()

HASH_LEAF_SIGNATURE = b"lh"
FAST_LEAF_SIGNATURE = b"lf"
LF_LH_SK_ELEMENT = Struct(
    "element_count" / Int16ul,
    "elements"
    / Array(
        this.element_count, Struct("key_node_offset" / Int32ul, "hash_value" / Int32ul)
    ),
).compile()

DIRTY_PAGES_REFERENCES = Struct("offset" / Int32ul, "size" / Int32ul)

TRANSACTION_LOG = Struct(
    "signature" / Const(b"HvLE"),
    "log_size" / Int32ul,
    "flags" / Int32ul,
    "sequence_number" / Int32ul,
    "hive_bin_size" / Int32ul,
    "dirty_pages_count" / Int32ul,
    "hash_1" / Int64ul,
    "hash_2" / Int64ul,
    "dirty_pages_references" / Array(this.dirty_pages_count, DIRTY_PAGES_REFERENCES),
)

BIG_DATA_SIGNATURE = b"db"
BIG_DATA_BLOCK = Struct(
    "signature" / Const(BIG_DATA_SIGNATURE),
    "number_of_segments" / Int16ul,
    "offset_to_list_of_segments" / Int32ul,
)

# This is the default name of a registry subkey
DEFAULT_VALUE = "(default)"

SECURITY_KEY_v1_1 = Struct(
    "unknown" * Bytes(4),
    "signature" / Const(b"sk"),
    "unknown" * Bytes(2),
    "prev_sk_offset" / Int32ul,
    "next_sk_offset" / Int32ul,
    "reference_count" / Int32ul,
    "security_descriptor_size" / Int32ul,
    "security_descriptor" / Bytes(this.security_descriptor_size),
)

SECURITY_KEY_v1_2 = Struct(
    "signature" / Const(b"sk"),
    "unknown" * Bytes(2),
    "prev_sk_offset" / Int32ul,
    "next_sk_offset" / Int32ul,
    "reference_count" / Int32ul,
    "security_descriptor_size" / Int32ul,
    "security_descriptor" / Bytes(this.security_descriptor_size),
    "ref_count" / Int32ul,
    "sdlen" / Int32ul,
)

# References:
# https://docs.microsoft.com/en-us/windows/win32/api/winnt/ns-winnt-security_descriptor
# https://docs.microsoft.com/en-us/windows/win32/secauthz/security-descriptor-control
SECURITY_DESCRIPTOR = Struct(
    "revision" / Bytes(1),
    "sbz1" / Bytes(1),
    "control"
    / FlagsEnum(
        Int16ul,
        SE_DACL_AUTO_INHERIT_REQ=0x0100,
        SE_DACL_AUTO_INHERITED=0x0400,
        SE_DACL_DEFAULTED=0x0008,
        SE_DACL_PRESENT=0x0004,
        SE_DACL_PROTECTED=0x1000,
        SE_GROUP_DEFAULTED=0x0002,
        SE_OWNER_DEFAULTED=0x0001,
        SE_RM_CONTROL_VALID=0x4000,
        SE_SACL_AUTO_INHERIT_REQ=0x0200,
        SE_SACL_AUTO_INHERITED=0x0800,
        SE_SACL_DEFAULTED=0x0008,
        SE_SACL_PRESENT=0x0010,
        SE_SACL_PROTECTED=0x2000,
        SE_SELF_RELATIVE=0x8000,
    ),
    "owner" / Int32ul,
    "group" / Int32ul,
    "offset_sacl" / Int32ul,
    "offset_dacl" / Int32ul,
)

# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-dtyp/c6ce4275-3d90-4890-ab3a-514745e4637e
# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-dtyp/f992ad60-0fe4-4b87-9fed-beb478836861
# https://flatcap.org/linux-ntfs/ntfs/attributes/security_descriptor.html
SID = Struct(
    "revision" / Int8ul,
    "sub_authority_count" / Int8ul,
    "identifier_authority" / Bytes(6),
    "subauthority" / Int32ul[this.sub_authority_count],
)

ACL = Struct(
    "revision" / Int8ul,
    "sbz1" * Int8ul,
    "acl_size" / Int16ul,
    "ace_count" / Int16ul,
    "sbz2" * Int16ul,
).compile()

ACE = Struct(
    "type"
    / Enum(
        Int8ul,
        ACCESS_ALLOWED=0,
        ACCESS_DENIED=1,
        SYSTEM_AUDIT=2,
        SYSTEM_ALARM=3,
        ACCESS_ALLOWED_COMPOUND=4,
        ACCESS_ALLOWED_OBJECT=5,
        ACCESS_DENIED_OBJECT=6,
        SYSTEM_AUDIT_OBJECT=7,
        SYSTEM_ALARM_OBJECT=8,
        ACCESS_ALLOWED_CALLBACK=9,
        ACCESS_DENIED_CALLBACK=10,
        ACCESS_ALLOWED_CALLBACK_OBJECT=11,
        ACCESS_DENIED_CALLBACK_OBJECT=12,
        SYSTEM_ALARM_CALLBACK=14,
        SYSTEM_AUDIT_CALLBACK_OBJECT=15,
        SYSTEM_ALARM_CALLBACK_OBJECT=16,
    ),
    "flags"
    / FlagsEnum(
        Int8ul,
        OBJECT_INHERIT_ACE=0x1,
        CONTAINER_INHERIT_ACE=0x2,
        NO_PROPAGATE_INHERIT_ACE=0x4,
        INHERIT_ONLY_ACE=0x8,
    ),
    "size" / Int16ul,
    "access_mask"
    / FlagsEnum(
        Int32ul,
        DELETE=0x00010000,
        READ_CONTROL=0x00020000,
        WRITE_DAC=0x00040000,
        WRITE_OWNER=0x00080000,
        SYNCHRONIZE=0x00100000,
        ACCESS_SYSTEM_SECURITY=0x01000000,
        MAXIMUM_ALLOWED=0x02000000,
        GENERIC_ALL=0x10000000,
        GENERIC_EXECUTE=0x20000000,
        GENERIC_WRITE=0x40000000,
        GENERIC_READ=0x80000000,
    ),
    "sid" / Bytes(this.size - 8),
).compile()



import logging

from regipy import RegistryKeyNotFoundException
from regipy.hive_types import NTUSER_HIVE_TYPE
from regipy.plugins.plugin import Plugin
from regipy.utils import convert_wintime

logger = logging.getLogger(__name__)

INSTALLED_SOFTWARE_PATH = r"\Software\Microsoft\Windows\CurrentVersion\Uninstall"


class InstalledProgramsNTUserPlugin(Plugin):
    NAME = "installed_programs_ntuser"
    DESCRIPTION = "Retrieve list of installed programs and their install date from the NTUSER Hive"
    COMPATIBLE_HIVE = NTUSER_HIVE_TYPE

    def _get_installed_software(self, subkey_path):
        try:
            uninstall_sk = self.registry_hive.get_key(subkey_path)
        except RegistryKeyNotFoundException as ex:
            logger.error(ex)
            return

        for installed_program in uninstall_sk.iter_subkeys():
            values = (
                {
                    x.name: x.value
                    for x in installed_program.iter_values(as_json=self.as_json)
                }
                if installed_program.values_count
                else {}
            )
            self.entries.append(
                {
                    "service_name": installed_program.name,
                    "timestamp": convert_wintime(
                        installed_program.header.last_modified, as_json=self.as_json
                    ),
                    "registry_path": subkey_path,
                    **values,
                }
            )

    def run(self):
        self._get_installed_software(INSTALLED_SOFTWARE_PATH)



import logging

from inflection import underscore

from regipy.exceptions import RegistryKeyNotFoundException
from regipy.hive_types import AMCACHE_HIVE_TYPE
from regipy.plugins.plugin import Plugin
from regipy.utils import convert_wintime

logger = logging.getLogger(__name__)

AMCACHE_FIELD_NUMERIC_MAPPINGS = {
    "0": "product_name",
    "1": "company_name",
    "2": "file_version_number",
    "3": "language_code",
    "4": "switchback_context",
    "5": "file_version",
    "6": "file_size",
    "7": "pe_header_hash",
    "8": "unknown1",
    "9": "pe_header_checksum",
    "a": "unknown2",
    "b": "unknown3",
    "c": "file_description",
    "d": "unknown4",
    "f": "linker_compile_time",
    "10": "unknown5",
    "11": "last_modified_timestamp",
    "12": "created_timestamp",
    "15": "full_path",
    "16": "unknown6",
    "17": "last_modified_timestamp_2",
    "100": "program_id",
    "101": "sha1",
}

WIN8_TS_FIELDS = [
    "last_modified_timestamp",
    "created_timestamp",
    "last_modified_timestamp_2",
]


class AmCachePlugin(Plugin):
    NAME = "amcache"
    DESCRIPTION = "Parse Amcache"
    COMPATIBLE_HIVE = AMCACHE_HIVE_TYPE

    def parse_amcache_file_entry(self, subkey):
        entry = {
            underscore(x.name): x.value
            for x in subkey.iter_values(as_json=self.as_json)
        }

        # Sometimes the value names might be numeric instead. Translate them:
        for k, v in AMCACHE_FIELD_NUMERIC_MAPPINGS.items():
            content = entry.pop(k, None)
            if content:
                entry[v] = content

        if "sha1" in entry:
            entry["sha1"] = entry["sha1"][4:]

        if "file_id" in entry and entry["file_id"] != 0:
            entry["file_id"] = entry["file_id"][4:]
            if "sha1" not in entry:
                entry["sha1"] = entry["file_id"]

        if "program_id" in entry:
            entry["program_id"] = entry["program_id"][4:]

        entry["timestamp"] = convert_wintime(
            subkey.header.last_modified, as_json=self.as_json
        )

        if "size" in entry:
            entry["size"] = (
                int(entry["size"], 16)
                if isinstance(entry["size"], str)
                else entry["size"]
            )

        is_pefile = entry.get("is_pe_file")
        if is_pefile is not None:
            entry["is_pe_file"] = bool(is_pefile)

        is_os_component = entry.get("is_os_component")
        if is_os_component is not None:
            entry["is_os_component"] = bool(is_os_component)

        if entry.get("link_date") == 0:
            entry.pop("link_date")

        for ts_field_name in WIN8_TS_FIELDS:
            ts = entry.pop(ts_field_name, None)
            if ts:
                entry[ts_field_name] = convert_wintime(ts, as_json=self.as_json)

        self.entries.append(entry)

    def run(self):
        logger.debug("Started AmCache Plugin...")

        try:
            amcache_file_subkey = self.registry_hive.get_key(r"\Root\File")
        except RegistryKeyNotFoundException:
            logger.error(r"Could not find \Root\File subkey")
            amcache_file_subkey = None

        try:
            amcache_inventory_file_subkey = self.registry_hive.get_key(
                r"\Root\InventoryApplicationFile"
            )
        except RegistryKeyNotFoundException:
            logger.info(r"Could not find \Root\InventoryApplicationFile subkey")
            amcache_inventory_file_subkey = None

        if amcache_file_subkey:
            for subkey in amcache_file_subkey.iter_subkeys():
                if subkey.header.subkey_count > 0:
                    for file_subkey in subkey.iter_subkeys():
                        self.parse_amcache_file_entry(file_subkey)
                if subkey.header.values_count > 0:
                    self.entries.append(subkey)

        if amcache_inventory_file_subkey:
            for file_subkey in amcache_inventory_file_subkey.iter_subkeys():
                self.parse_amcache_file_entry(file_subkey)



import logging

from regipy.hive_types import SYSTEM_HIVE_TYPE
from regipy.plugins.plugin import Plugin
from regipy.plugins.system.external.ShimCacheParser import get_shimcache_entries

logger = logging.getLogger(__name__)

COMPUTER_NAME_PATH = r"Control\Session Manager"


class ShimCachePlugin(Plugin):
    NAME = "shimcache"
    DESCRIPTION = "Parse Shimcache artifact"
    COMPATIBLE_HIVE = SYSTEM_HIVE_TYPE

    def run(self):
        logger.debug("Started Shim Cache Plugin...")

        for subkey_path in self.registry_hive.get_control_sets(COMPUTER_NAME_PATH):
            appcompat_cache = self.registry_hive.get_key(subkey_path).get_subkey(
                "AppCompatCache"
            )

            if not appcompat_cache:
                logger.info("No shimcache data was found")
                return

            shimcache = appcompat_cache.get_value("AppCompatCache")
            if shimcache:
                for entry in get_shimcache_entries(shimcache, as_json=self.as_json):
                    self.entries.append(entry)



import logging
from inflection import underscore

from regipy import RegistryKeyNotFoundException, convert_wintime
from regipy.hive_types import NTUSER_HIVE_TYPE
from regipy.plugins.plugin import Plugin

logger = logging.getLogger(__name__)


TYPED_URLS_KEY_PATH = r"\Software\Microsoft\Internet Explorer\TypedURLs"


class TypedUrlsPlugin(Plugin):
    NAME = "typed_urls"
    DESCRIPTION = "Retrieve the typed URLs from IE history"
    COMPATIBLE_HIVE = NTUSER_HIVE_TYPE

    def run(self):
        try:
            subkey = self.registry_hive.get_key(TYPED_URLS_KEY_PATH)
        except RegistryKeyNotFoundException as ex:
            logger.error(
                f"Could not find {self.NAME} plugin data at: {TYPED_URLS_KEY_PATH}: {ex}"
            )
            return None

        self.entries = {
            "last_write": convert_wintime(
                subkey.header.last_modified, as_json=self.as_json
            ),
            "entries": [
                {underscore(x.name): x.value}
                for x in subkey.iter_values(as_json=self.as_json)
            ],
        }



import logging

from regipy.hive_types import SOFTWARE_HIVE_TYPE
from regipy.plugins.plugin import Plugin
from regipy.utils import convert_wintime

logger = logging.getLogger(__name__)

IMAGE_FILE_EXECUTION_OPTIONS = (
    r"\Microsoft\Windows NT\CurrentVersion\Image File Execution Options"
)


class ImageFileExecutionOptions(Plugin):
    NAME = "image_file_execution_options"
    DESCRIPTION = "Retrieve image file execution options - a persistence method"
    COMPATIBLE_HIVE = SOFTWARE_HIVE_TYPE

    def run(self):
        image_file_execution_options = self.registry_hive.get_key(
            IMAGE_FILE_EXECUTION_OPTIONS
        )
        if image_file_execution_options.subkey_count:
            for subkey in image_file_execution_options.iter_subkeys():
                values = (
                    {x.name: x.value for x in subkey.iter_values(as_json=self.as_json)}
                    if subkey.values_count
                    else {}
                )
                self.entries.append(
                    {
                        "name": subkey.name,
                        "timestamp": convert_wintime(
                            subkey.header.last_modified, as_json=self.as_json
                        ),
                        **values,
                    }
                )



import logging

from regipy.hive_types import SOFTWARE_HIVE_TYPE
from regipy.plugins.plugin import Plugin
from regipy.utils import convert_wintime
from regipy.exceptions import RegistryKeyNotFoundException

logger = logging.getLogger(__name__)


WIN_VER_PATH = r"\Microsoft\Windows\CurrentVersion\WindowsUpdate"
value_list = ("LastRestorePointSetTime", "SusClientId")


class SusclientPlugin(Plugin):
    NAME = "susclient_plugin"
    DESCRIPTION = "Extracts SusClient* info, including HDD SN"
    COMPATIBLE_HIVE = SOFTWARE_HIVE_TYPE

    def can_run(self):
        return self.registry_hive.hive_type == SOFTWARE_HIVE_TYPE

    def run(self):
        logger.info("Started susclient Plugin...")

        try:
            key = self.registry_hive.get_key(WIN_VER_PATH)
        except RegistryKeyNotFoundException as ex:
            logger.error(f"Could not find {self.NAME} subkey at {WIN_VER_PATH}: {ex}")
            return None

        self.entries = {
            WIN_VER_PATH: {
                "last_write": convert_wintime(key.header.last_modified).isoformat()
            }
        }

        for val in key.iter_values():
            if val.name == "SusClientIdValidation":
                self.entries[WIN_VER_PATH][val.name] = get_SN(val.value)
            elif val.name in value_list:
                self.entries[WIN_VER_PATH][val.name] = val.value


def get_SN(data):
    offset = int(data[:2], 16)
    length = int(data[4:6], 16)
    return bytes.fromhex(data[2 * offset: 2 * (length + offset)]).decode("utf-16le")



import logging

from regipy.hive_types import SOFTWARE_HIVE_TYPE
from regipy.plugins.plugin import Plugin
from regipy.utils import convert_wintime
from regipy.exceptions import RegistryKeyNotFoundException

logger = logging.getLogger(__name__)


SYS_RESTORE_PATH = r"\Microsoft\Windows NT\CurrentVersion\SystemRestore"
value_list = "DisableSR"


class DisableSRPlugin(Plugin):
    NAME = "disablesr_plugin"
    DESCRIPTION = "Gets the value that turns System Restore either on or off"
    COMPATIBLE_HIVE = SOFTWARE_HIVE_TYPE

    def can_run(self):
        return self.registry_hive.hive_type == SOFTWARE_HIVE_TYPE

    def run(self):
        logger.info("Started disablesr Plugin...")

        try:
            key = self.registry_hive.get_key(SYS_RESTORE_PATH)
        except RegistryKeyNotFoundException as ex:
            logger.error(
                f"Could not find {self.NAME} subkey at {SYS_RESTORE_PATH}: {ex}"
            )
            return None

        self.entries = {
            SYS_RESTORE_PATH: {
                "last_write": convert_wintime(key.header.last_modified).isoformat()
            }
        }

        for val in key.iter_values():
            if val.name in value_list:
                self.entries[SYS_RESTORE_PATH][val.name] = val.value



import logging

from regipy.exceptions import RegistryValueNotFoundException
from regipy.hive_types import SYSTEM_HIVE_TYPE
from regipy.plugins.plugin import Plugin
from regipy.utils import convert_wintime

logger = logging.getLogger(__name__)

WDIGEST_PATH = r"Control\SecurityProviders\WDigest"


class WDIGESTPlugin(Plugin):
    NAME = "wdigest"
    DESCRIPTION = "Get WDIGEST configuration"
    COMPATIBLE_HIVE = SYSTEM_HIVE_TYPE

    def run(self):
        logger.debug("Started WDIGEST Plugin...")
        for subkey_path in self.registry_hive.get_control_sets(WDIGEST_PATH):
            subkey = self.registry_hive.get_key(subkey_path)

            try:
                self.entries.append(
                    {
                        "subkey": subkey_path,
                        "use_logon_credential": subkey.get_value(
                            "UseLogonCredential", as_json=self.as_json
                        ),
                        "timestamp": convert_wintime(
                            subkey.header.last_modified, as_json=self.as_json
                        ),
                    }
                )
            except RegistryValueNotFoundException as ex:
                logger.exception(f"Could not find Wdigest for registry hive: {ex}")
                continue



import binascii
import datetime as dt

from typing import List, Optional, Union

import logging

import attr

from construct import (
    Bytes,
    CString,
    EnumIntegerString,
    GreedyRange,
    Int32sl,
    Int32ul,
    Int64ul,
    StreamError,
    ConstError,
)
from io import BytesIO


from regipy.exceptions import (
    NoRegistrySubkeysException,
    RegistryKeyNotFoundException,
    RegistryValueNotFoundException,
    RegipyGeneralException,
    UnidentifiedHiveException,
    RegistryParsingException,
)
from regipy.hive_types import SUPPORTED_HIVE_TYPES
from regipy.security_utils import convert_sid, get_acls
from regipy.structs import (
    REGF_HEADER,
    HBIN_HEADER,
    CM_KEY_NODE,
    LF_LH_SK_ELEMENT,
    VALUE_KEY,
    INDEX_ROOT,
    REGF_HEADER_SIZE,
    INDEX_ROOT_SIGNATURE,
    LEAF_INDEX_SIGNATURE,
    FAST_LEAF_SIGNATURE,
    HASH_LEAF_SIGNATURE,
    BIG_DATA_BLOCK,
    INDEX_LEAF,
    DEFAULT_VALUE,
    SECURITY_KEY_v1_1,
    SECURITY_DESCRIPTOR,
    SID,
    VALUE_TYPE_ENUM,
)
from regipy.utils import (
    boomerang_stream,
    convert_wintime,
    identify_hive_type,
    MAX_LEN,
    trim_registry_data_for_error_msg,
    try_decode_binary,
)

logger = logging.getLogger(__name__)


@attr.s
class Cell:
    """
    Represents a Registry cell header
    """

    offset: int = attr.ib()
    cell_type: str = attr.ib()
    size: int = attr.ib()


@attr.s
class VKRecord:
    """
    The VK Record contains a value
    """

    value_type: EnumIntegerString = attr.ib()
    value_type_str: str = attr.ib()
    value: bytes = attr.ib()
    size: int = attr.ib(default=0)
    is_corrupted: bool = attr.ib(default=False)


@attr.s
class LIRecord:
    data: bytes = attr.ib()


@attr.s
class Value:
    name: str = attr.ib()
    value: Union[str, int, bytes] = attr.ib()
    value_type: str = attr.ib()
    is_corrupted: bool = attr.ib(default=False)


@attr.s
class Subkey:
    subkey_name: str = attr.ib()
    path: str = attr.ib()
    timestamp: dt.datetime = attr.ib()
    values_count: int = attr.ib()
    values: List[Value] = attr.ib(factory=list)

    # This field will be used if a partial hive was given, if not it would be None.
    actual_path: Optional[str] = attr.ib(default=None)


class RIRecord:
    data = None
    header = None

    def __init__(self, stream):
        self.header = INDEX_ROOT.parse_stream(stream)


class RegistryHive:
    CONTROL_SETS = [r"\ControlSet001", r"\ControlSet002"]

    def __init__(self, hive_path, hive_type=None, partial_hive_path=None):
        """
        Represents a registry hive
        :param hive_path: Path to the registry hive
        :param hive_type: The hive type can be specified if this is a partial hive,
                          or for some other reason regipy cannot identify the hive type
        :param partial_hive_path: The path from which the partial hive actually starts, for example:
                                  hive_type=ntuser partial_hive_path="/Software" would mean
                                  this is actually a HKCU hive, starting from HKCU/Software
        """

        self.partial_hive_path = None
        self.hive_type = None

        with open(hive_path, "rb") as f:
            self._stream = BytesIO(f.read())

        with boomerang_stream(self._stream) as s:
            self.header = REGF_HEADER.parse_stream(s)

            # Get the first cell in root HBin, which is the root NKRecord:
            root_hbin = self.get_hbin_at_offset()
            root_hbin_cell = next(root_hbin.iter_cells(s))
            self.root = NKRecord(root_hbin_cell, s)
        self.name = self.header.file_name

        if hive_type:
            if hive_type.lower() in SUPPORTED_HIVE_TYPES:
                self.hive_type = hive_type
            else:
                raise UnidentifiedHiveException(
                    f"{hive_type} is not a supported hive type: "
                    f"only the following are supported: {SUPPORTED_HIVE_TYPES}"
                )
        else:
            try:
                self.hive_type = identify_hive_type(self.name)
            except UnidentifiedHiveException:
                logger.info(
                    f"Hive type for {hive_path} was not identified: {self.name}"
                )

        if partial_hive_path:
            self.partial_hive_path = partial_hive_path

    def recurse_subkeys(
        self,
        nk_record=None,
        path_root=None,
        as_json=False,
        is_init=True,
        fetch_values=True,
    ):
        """
        Recurse over a subkey, and yield all of its subkeys and values
        :param nk_record: an instance of NKRecord from which to start iterating, if None, will start from Root
        :param path_root: If we are iterating an incomplete hive, for example a hive tree starting
                          from ControlSet001 and not SYSTEM, there is no way to know that.
                          This string will be added as prefix to all paths.
        :param as_json: Whether to normalize the data as JSON or not
        :param fetch_values: If False, subkey values will not be returned, but the iteration will be faster
        """
        # If None, will start iterating from Root NK entry
        if not nk_record:
            nk_record = self.root

        # Iterate over subkeys
        if nk_record.header.subkey_count:
            for subkey in nk_record.iter_subkeys():
                if path_root:
                    subkey_path = (
                        r"{}\{}".format(path_root, subkey.name)
                        if path_root
                        else r"\{}".format(subkey.name)
                    )
                else:
                    subkey_path = f"\\{subkey.name}"

                # Leaf Index records do not contain subkeys
                if isinstance(subkey, LIRecord):
                    continue

                if subkey.subkey_count:
                    yield from self.recurse_subkeys(
                        nk_record=subkey,
                        path_root=subkey_path,
                        as_json=as_json,
                        is_init=False,
                        fetch_values=fetch_values,
                    )

                values = []
                if fetch_values:
                    if subkey.values_count:
                        try:
                            if as_json:
                                values = [
                                    attr.asdict(x)
                                    for x in subkey.iter_values(as_json=as_json)
                                ]
                            else:
                                values = list(subkey.iter_values(as_json=as_json))
                        except RegistryParsingException:
                            logger.exception(
                                f"Failed to parse hive value at path: {trim_registry_data_for_error_msg(path_root)}"
                            )

                ts = convert_wintime(subkey.header.last_modified)
                yield Subkey(
                    subkey_name=subkey.name,
                    path=subkey_path,
                    timestamp=ts.isoformat() if as_json else ts,
                    values=values,
                    values_count=subkey.values_count,
                    actual_path=(
                        f"{self.partial_hive_path}{subkey_path}"
                        if self.partial_hive_path
                        else None
                    ),
                )

        if is_init:
            # Get the values of the subkey
            values = []
            if nk_record.values_count:
                try:
                    if as_json:
                        values = [
                            attr.asdict(x)
                            for x in nk_record.iter_values(as_json=as_json)
                        ]
                    else:
                        values = list(nk_record.iter_values(as_json=as_json))
                except RegistryParsingException as ex:
                    logger.exception(
                        f"Failed to parse hive value at path: {trim_registry_data_for_error_msg(path_root)}: {ex}"
                    )
                    values = []

            ts = convert_wintime(nk_record.header.last_modified)
            subkey_path = path_root or "\\"
            yield Subkey(
                subkey_name=nk_record.name,
                path=subkey_path,
                timestamp=ts.isoformat() if as_json else ts,
                values=values,
                values_count=len(values),
                actual_path=(
                    f"{self.partial_hive_path}\\{subkey_path}"
                    if self.partial_hive_path
                    else None
                ),
            )

    def get_hbin_at_offset(self, offset=0):
        """
        This offset is from start of data (meaning that it starts from 4096)
        If not offset is given, will return the root Hbin
        :return:
        """
        self._stream.seek(REGF_HEADER_SIZE + offset)
        return HBin(self._stream)

    def get_key(self, key_path):
        if self.partial_hive_path:
            if key_path.startswith(self.partial_hive_path):
                key_path = key_path.partition(self.partial_hive_path)[-1]
            else:
                raise RegistryKeyNotFoundException(
                    f"Did not find subkey at {key_path}, because this is a partial hive"
                )

        logger.debug("Getting key: {}".format(key_path))

        # If the key path is \ we are just refering to root
        if key_path == "\\":
            return self.root

        # If the path contain slashes, this is a full path. Split it
        if "\\" in key_path:
            key_path_parts = key_path.split("\\")[1:]
        else:
            key_path_parts = [key_path]

        previous_key_name = []

        subkey = self.root.get_subkey(key_path_parts.pop(0), raise_on_missing=False)

        if not subkey:
            raise RegistryKeyNotFoundException(
                "Did not find subkey at {}".format(key_path)
            )

        if not key_path_parts:
            return subkey

        for path_part in key_path_parts:
            new_path = "\\".join(previous_key_name)
            previous_key_name.append(subkey.name)
            subkey = subkey.get_subkey(path_part, raise_on_missing=False)

            if not subkey:
                raise RegistryKeyNotFoundException(
                    "Did not find {} at {}".format(path_part, new_path)
                )
        return subkey

    def get_control_sets(self, registry_path):
        """
        Get the optional control sets for a registry hive
        :param registry_path:
        :return: A list of paths, including the control sets
        """

        found_control_sets = []
        for cs in self.CONTROL_SETS:
            try:
                found_control_sets.append(self.get_key(cs))
            except RegistryKeyNotFoundException:
                continue
        result = [
            r"\{}\{}".format(subkey.name, registry_path)
            for subkey in found_control_sets
        ]
        logger.debug("Found control sets: {}".format(result))
        return result


class HBin:
    def __init__(self, stream):
        """
        :param stream: a stream at the start of the hbin block
        """
        self.header = HBIN_HEADER.parse_stream(stream)
        self.hbin_data_offset = stream.tell()

    def iter_cells(self, stream):
        stream.seek(self.hbin_data_offset)
        offset = stream.tell()
        while offset < self.hbin_data_offset + self.header.size - HBIN_HEADER.sizeof():
            hbin_cell_size = Int32sl.parse_stream(stream)

            # If the cell size is positive, it means it is unallocated. We are not interested in those on a regular run
            if hbin_cell_size >= 0:
                continue

            bytes_to_read = (hbin_cell_size * -1) - 4

            cell_type = Bytes(2).parse_stream(stream)

            # Yield the cell
            yield Cell(
                cell_type=cell_type.decode(), offset=stream.tell(), size=bytes_to_read
            )

            # Go to the next cell
            offset += stream.tell() + bytes_to_read


class NKRecord:
    """
    The NKRecord represents a Name Key entry
    """

    def __init__(self, cell, stream):
        stream.seek(cell.offset)
        self.header = CM_KEY_NODE.parse_stream(stream)
        self._stream = stream

        # Sometimes the key names are ASCII and sometimes UTF-16 little endian
        if self.header.flags.KEY_COMP_NAME:
            # Compressed (ASCII) key name
            self.name = self.header.key_name_string.decode("ascii", errors="replace")
        else:
            # Unicode (UTF-16) key name
            self.name = self.header.key_name_string.decode(
                "utf-16-le", errors="replace"
            )
            logger.debug(f'Unicode key name identified: "{self.name}"')

        self.subkey_count = self.header.subkey_count
        self.values_count = self.header.values_count
        self.volatile_subkeys_count = self.header.volatile_subkey_count

    def get_subkey(self, key_name, raise_on_missing=True):
        if not self.subkey_count and raise_on_missing:
            raise NoRegistrySubkeysException(
                "No subkeys for {}".format(self.header.key_name_string)
            )

        for subkey in self.iter_subkeys():
            # This should not happen
            if not isinstance(subkey, NKRecord):
                raise RegipyGeneralException(f"Unknown record type: {subkey}")

            if subkey.name.upper() == key_name.upper():
                return subkey

        if raise_on_missing:
            raise NoRegistrySubkeysException(
                "No subkey {} for {}".format(key_name, self.header.key_name_string)
            )

    def iter_subkeys(self):

        if not self.header.subkey_count:
            return None

        # Go to the offset where the subkey list starts (+4 is because of the cell header)
        target_offset = REGF_HEADER_SIZE + 4 + self.header.subkeys_list_offset
        self._stream.seek(target_offset)

        # Read the signature
        try:
            signature = Bytes(2).parse_stream(self._stream)
        except StreamError as ex:
            raise RegistryParsingException(
                f"Bad subkey at offset {target_offset}: {ex}"
            )

        # LF,  LH and RI contain subkeys
        if signature in [
            HASH_LEAF_SIGNATURE,
            FAST_LEAF_SIGNATURE,
            LEAF_INDEX_SIGNATURE,
        ]:
            yield from self._parse_subkeys(self._stream, signature=signature)
        # RI contains pointers to arrays of subkeys
        elif signature == INDEX_ROOT_SIGNATURE:
            ri_record = RIRecord(self._stream)
            if ri_record.header.element_count > 0:
                for element in ri_record.header.elements:
                    # We skip 6 because of the signature as well as the cell header
                    element_target_offset = (
                        REGF_HEADER_SIZE + 4 + element.subkey_list_offset
                    )
                    self._stream.seek(element_target_offset)
                    yield from self._parse_subkeys(self._stream)

    @staticmethod
    def _parse_subkeys(stream, signature=None):
        """
        Parse an LI , LF or LH Record
        :param stream: A stream at the header of the LH or LF entry, skipping the signature
        :return:
        """
        if not signature:
            signature = stream.read(2)

        if signature in [HASH_LEAF_SIGNATURE, FAST_LEAF_SIGNATURE]:
            subkeys = LF_LH_SK_ELEMENT.parse_stream(stream)
        elif signature == LEAF_INDEX_SIGNATURE:
            subkeys = INDEX_LEAF.parse_stream(stream)
        else:
            raise RegistryParsingException(
                f"Expected a known signature, got: {signature} at offset {stream.tell()}"
            )

        for subkey in subkeys.elements:
            stream.seek(REGF_HEADER_SIZE + subkey.key_node_offset)

            # This cell should always be allocated, therefor we expect a negative size
            cell_size = Int32sl.parse_stream(stream) * -1

            # We read to this offset and skip 2 bytes, because that is the cell size we just read
            nk_cell = Cell(cell_type="nk", offset=stream.tell() + 2, size=cell_size)
            nk_record = NKRecord(cell=nk_cell, stream=stream)
            yield nk_record

    @staticmethod
    def read_value(vk, stream):
        """
        Read a registry value
        :param vk: A parse VK record
        :param stream: The registry stream
        :return: A VKRecord
        """
        stream.seek(REGF_HEADER_SIZE + 4 + vk.data_offset)
        data_type = vk.data_type
        data = stream.read(vk.data_size)
        return VKRecord(
            value_type=data_type,
            value_type_str=str(data_type),
            value=data,
            size=vk.data_size,
        )

    @staticmethod
    def _parse_indirect_block(stream, value):
        # This is an indirect datablock (Bigger than 16344, therefor we handle it differently)
        # The value inside the vk entry actually contains a pointer to the buffers containing the data
        big_data_block_header = BIG_DATA_BLOCK.parse(value.value)

        # Go to the start of the segment offset list
        stream.seek(REGF_HEADER_SIZE + big_data_block_header.offset_to_list_of_segments)
        buffer = BytesIO()

        # Read them sequentially until we got all the size of the VK
        value_size = value.size
        while value_size > 0:
            data_segment_offset = Int32ul.parse_stream(stream)
            with boomerang_stream(stream) as tmpstream:
                tmpstream.seek(REGF_HEADER_SIZE + 4 + data_segment_offset)
                tmpbuffer = tmpstream.read(min(0x3FD8, value_size))
                value_size -= len(tmpbuffer)
                buffer.write(tmpbuffer)
        buffer.seek(0)
        return buffer.read()

    def iter_values(self, as_json=False, max_len=MAX_LEN, trim_values=True):
        """
        Get the values of a subkey. Will raise if no values exist
        :param as_json: Whether to normalize the data as JSON or not
        :param max_len: Max length of value to return
        :param trim_values: whether to trim values to MAX_LEN
        :return: List of values for the subkey
        """
        if not self.values_count:
            return

        # Get the offset of the values key. We skip 4 because of Cell Header
        target_offset = REGF_HEADER_SIZE + 4 + self.header.values_list_offset
        self._stream.seek(target_offset)

        for _ in range(self.values_count):
            is_corrupted = False
            try:
                vk_offset = Int32ul.parse_stream(self._stream)
            except StreamError:
                logger.info(f"Skipping bad registry VK at {self._stream.tell()}")
                raise RegistryParsingException(
                    f"Bad registry VK at {self._stream.tell()}"
                )

            with boomerang_stream(self._stream) as substream:
                actual_vk_offset = REGF_HEADER_SIZE + 4 + vk_offset
                substream.seek(actual_vk_offset)
                try:
                    vk = VALUE_KEY.parse_stream(substream)
                except (ConstError, StreamError):
                    logger.error(
                        f"Could not parse VK at {substream.tell()}, registry hive is probably corrupted."
                    )
                    return

                value = self.read_value(vk, substream)

                if vk.name_size == 0:
                    value_name = "(default)"
                elif vk.flags.VALUE_COMP_NAME:
                    # Compressed (ASCII) value name
                    value_name = vk.name.decode("ascii", errors="replace")
                else:
                    # Unicode (UTF-16) value name
                    value_name = vk.name.decode("utf-16-le", errors="replace")
                    logger.debug(f'Unicode value name identified: "{value_name}"')

                # If the value is bigger than this value, it means this is a DEVPROP structure
                # https://doxygen.reactos.org/d0/dba/devpropdef_8h_source.html
                # https://sourceforge.net/p/mingw-w64/mingw-w64/ci/668a1d3e85042c409e0c292e621b3dc0aa26177c/tree/
                # mingw-w64-headers/include/devpropdef.h?diff=dd86a3b7594dadeef9d6a37c4b6be3ca42ef7e94
                # We currently do not support these, We are going to make the best effort to dump as string.
                # This int casting will always work because the data_type is construct's EnumIntegerString
                if int(vk.data_type) > 0xFFFF0000:
                    data_type = VALUE_TYPE_ENUM.parse(
                        Int32ul.build(int(vk.data_type) & 0xFFFF)
                    )
                    logger.info(
                        f"Value at {hex(actual_vk_offset)} contains DEVPROP structure of type {data_type}"
                    )

                # Skip this unknown data type, research pending :)
                # TODO: Add actual parsing
                elif int(vk.data_type) == 0x200000:
                    logger.info(
                        f"Skipped unknown data type value at {actual_vk_offset}"
                    )
                    continue
                else:
                    data_type = str(vk.data_type)

                if data_type in ["REG_SZ", "REG_EXPAND", "REG_EXPAND_SZ"]:
                    if vk.data_size >= 0x80000000:
                        # data is contained in the data_offset field
                        value.size -= 0x80000000
                        actual_value = vk.data_offset
                    elif vk.data_size > 0x3FD8 and value.value[:2] == b"db":
                        data = self._parse_indirect_block(substream, value)
                        actual_value = try_decode_binary(
                            data, as_json=as_json, trim_values=trim_values
                        )
                    else:
                        actual_value = try_decode_binary(
                            value.value, as_json=as_json, trim_values=trim_values
                        )
                elif data_type in ["REG_BINARY", "REG_NONE"]:
                    if vk.data_size >= 0x80000000:
                        # data is contained in the data_offset field
                        actual_value = vk.data_offset
                    elif vk.data_size > 0x3FD8 and value.value[:2] == b"db":
                        try:
                            actual_value = self._parse_indirect_block(substream, value)

                            actual_value = (
                                try_decode_binary(
                                    actual_value, as_json=True, trim_values=trim_values
                                )
                                if as_json
                                else actual_value
                            )
                        except ConstError:
                            logger.error(f"Bad value at {actual_vk_offset}")
                            continue
                    else:
                        # Return the actual data
                        actual_value = (
                            binascii.b2a_hex(value.value).decode()[:max_len]
                            if trim_values
                            else value.value
                        )
                elif data_type == "REG_SZ":
                    actual_value = try_decode_binary(
                        value.value, as_json=as_json, trim_values=trim_values
                    )
                elif data_type == "REG_DWORD":
                    # If the data size is bigger than 0x80000000, data is actually stored in the VK data offset.
                    actual_value = (
                        vk.data_offset
                        if vk.data_size >= 0x80000000
                        else Int32ul.parse(value.value)
                    )
                elif data_type == "REG_QWORD":
                    actual_value = (
                        vk.data_offset
                        if vk.data_size >= 0x80000000
                        else Int64ul.parse(value.value)
                    )
                elif data_type == "REG_MULTI_SZ":
                    parsed_value = GreedyRange(CString("utf-16-le")).parse(value.value)
                    # Because the ListContainer object returned by Construct cannot be turned into a list,
                    # we do this trick
                    actual_value = [x for x in parsed_value if x]
                # We currently dumps this as hex string or raw
                # TODO: Add actual parsing
                elif data_type in [
                    "REG_RESOURCE_REQUIREMENTS_LIST",
                    "REG_RESOURCE_LIST",
                ]:
                    actual_value = (
                        binascii.b2a_hex(value.value).decode()[:max_len]
                        if trim_values
                        else value.value
                    )
                elif data_type == "REG_FILETIME":
                    actual_value = convert_wintime(
                        Int64ul.parse(value.value), as_json=as_json
                    )
                else:
                    actual_value = try_decode_binary(
                        value.value, as_json=as_json, trim_values=trim_values
                    )
                yield Value(
                    name=value_name,
                    value_type=data_type,
                    value=actual_value,
                    is_corrupted=is_corrupted,
                )

    def get_value(
        self,
        value_name=DEFAULT_VALUE,
        as_json=False,
        raise_on_missing=False,
        case_sensitive=True,
    ):
        """
        Get a value by name. Will raise if raise_on_missing is set,
        if no value name is given, will return the content of the default value
        :param value_name: The value name to look for
        :param as_json: Whether to normalize the data as JSON or not
        :param raise_on_missing: Will raise exception if value is missing, else will return None
        :return:
        """
        value_name = value_name if case_sensitive else value_name.lower()
        for value in self.iter_values(as_json=as_json, trim_values=False):
            v = value.name if case_sensitive else value.name.lower()
            if v == value_name:
                return value.value

        if raise_on_missing:
            raise RegistryValueNotFoundException(
                "Did not find the value {} on subkey {}".format(value_name, self.name)
            )
        return None

    def get_values(self, as_json=False, trim_values=False):
        return [x for x in self.iter_values(as_json=as_json, trim_values=trim_values)]

    def get_security_key_info(self):
        self._stream.seek(REGF_HEADER_SIZE + self.header.security_key_offset)
        # TODO: If parsing fails, parse with SECURITY_KEY_v1_2
        security_key = SECURITY_KEY_v1_1.parse_stream(self._stream)
        security_descriptor = SECURITY_DESCRIPTOR.parse(
            security_key.security_descriptor
        )

        with boomerang_stream(self._stream) as s:
            security_base_offset = (
                REGF_HEADER_SIZE + self.header.security_key_offset + 24
            )

            s.seek(security_base_offset + security_descriptor.owner)
            owner_sid = convert_sid(SID.parse_stream(s))

            s.seek(security_base_offset + security_descriptor.group)
            group_sid = convert_sid(SID.parse_stream(s))

            sacl_aces = None
            if security_descriptor.offset_sacl > 0:
                s.seek(security_base_offset + security_descriptor.offset_sacl)
                sacl_aces = get_acls(s)

            dacl_aces = None
            if security_descriptor.offset_dacl > 0:
                s.seek(security_base_offset + security_descriptor.offset_dacl)
                dacl_aces = get_acls(s)
            return {
                "owner": owner_sid,
                "group": group_sid,
                "dacl": dacl_aces,
                "sacl": sacl_aces,
            }

    def get_class_name(self) -> str:
        """
        Gets the key class name as would be returned via
        the `lpClass` argument of the `RegQueryInfoKey()` function.
        """

        # Get the offset of the class name string. We skip 4 because of Cell Header
        read_offset = REGF_HEADER_SIZE + 4 + self.header.class_name_offset

        self._stream.seek(read_offset)
        class_name = self._stream.read(self.header.class_name_size)

        return class_name.decode("utf-16-le", errors="replace")

    def __dict__(self):
        return {
            "name": self.name,
            "subkey_count": self.subkey_count,
            "value_count": self.values_count,
            "values": (
                {x["name"]: x["value"] for x in self.iter_values()}
                if self.values_count
                else None
            ),
            "subkeys": (
                {x["name"] for x in self.iter_subkeys()} if self.subkey_count else None
            ),
            "timestamp": convert_wintime(self.header.last_modified, as_json=True),
            "volatile_subkeys": self.volatile_subkeys_count,
        }



import logging

from regipy import (
    RegistryKeyNotFoundException,
    convert_wintime,
    NoRegistrySubkeysException,
)
from regipy.hive_types import NTUSER_HIVE_TYPE
from regipy.plugins.plugin import Plugin

logger = logging.getLogger(__name__)

TSCLIENT_HISTORY_PATH = r"\Software\Microsoft\Terminal Server Client\Servers"


class TSClientPlugin(Plugin):
    NAME = "terminal_services_history"
    DESCRIPTION = "Retrieve history of RDP connections"
    COMPATIBLE_HIVE = NTUSER_HIVE_TYPE

    def run(self):
        try:
            tsclient_subkey = self.registry_hive.get_key(TSCLIENT_HISTORY_PATH)
        except (RegistryKeyNotFoundException, NoRegistrySubkeysException) as ex:
            logger.error(ex)
            return

        for server in tsclient_subkey.iter_subkeys():
            self.entries.append(
                {
                    "server": server.name,
                    "last_connection": convert_wintime(
                        server.header.last_modified, as_json=self.as_json
                    ),
                    "username_hint": server.get_value("UsernameHint"),
                }
            )



import logging

from regipy.hive_types import SYSTEM_HIVE_TYPE
from regipy.plugins.plugin import Plugin
from regipy.exceptions import RegistryKeyNotFoundException
from regipy.utils import convert_wintime
from regipy.utils import convert_filetime2


logger = logging.getLogger(__name__)

SHUTDOWN_DATA_PATH = r"Control\Windows"


class ShutdownPlugin(Plugin):
    NAME = "shutdown"
    DESCRIPTION = "Get shutdown data"
    COMPATIBLE_HIVE = SYSTEM_HIVE_TYPE

    def run(self):
        self.entries = {}
        shutdown_subkeys = self.registry_hive.get_control_sets(SHUTDOWN_DATA_PATH)
        for shutdown_subkey in shutdown_subkeys:
            try:
                shutdown = self.registry_hive.get_key(shutdown_subkey)
            except RegistryKeyNotFoundException as ex:
                logger.error(
                    f"Could not find {self.NAME} subkey at {shutdown_subkey}: {ex}"
                )
                continue
            self.entries[shutdown_subkey] = {
                "last_write": convert_wintime(shutdown.header.last_modified).isoformat()
            }
            for val in shutdown.iter_values():
                if val.name == "ShutdownTime":
                    self.entries[shutdown_subkey]["date"] = convert_filetime2(val.value)



import codecs

import logging

from construct import Bytes, Const, ConstError, Int32ul, Int64ul, Struct

from regipy.exceptions import RegistryKeyNotFoundException
from regipy.hive_types import NTUSER_HIVE_TYPE
from regipy.plugins.plugin import Plugin
from regipy.utils import convert_wintime

logger = logging.getLogger(__name__)

USER_ASSIST_KEY_PATH = r"\Software\Microsoft\Windows\CurrentVersion\Explorer\UserAssist"

# guids for the various Operating Systems
GUIDS = [
    "{75048700-EF1F-11D0-9888-006097DEACF9}",  # Windows XP GUIDs
    "{5E6AB780-7743-11CF-A12B-00AA004AE837}",
    "{75048700-EF1F-11D0-9888-006097DEACF9}",  # Windows vista
    "{5E6AB780-7743-11CF-A12B-00AA004AE837}",
    "{F4E57C4B-2036-45F0-A9AB-443BCFE33D9F}",  # Windows 7
    "{CEBFF5CD-ACE2-4F4F-9178-9926F41749EA}",
    "{FA99DFC7-6AC2-453A-A5E2-5E2AFF4507BD}",  # Windows 8
    "{F4E57C4B-2036-45F0-A9AB-443BCFE33D9F}",
    "{F2A1CB5A-E3CC-4A2E-AF9D-505A7009D442}",
    "{CEBFF5CD-ACE2-4F4F-9178-9926F41749EA}",
    "{CAA59E3C-4792-41A5-9909-6A6A8D32490E}",
    "{B267E3AD-A825-4A09-82B9-EEC22AA3B847}",
    "{A3D53349-6E61-4557-8FC7-0028EDCEEBF6}",
    "{9E04CAB2-CC14-11DF-BB8C-A2F1DED72085}",
]

GUID_TO_PATH_MAPPINGS = {
    "{1AC14E77-02E7-4E5D-B744-2EB1AE5198B7}": r"%SYSTEM32%",
    "{6D809377-6AF0-444B-8957-A3773F02200E}": r"%PROGRAMFILES%",
    "{7C5A40EF-A0FB-4BFC-874A-C0F2E0B9FA8E}": r"%PROGRAMFILES(X86)%",
    "{F38BF404-1D43-42F2-9305-67DE0B28FC23}": r"%WINDIR%",
    "{0139D44E-6AFE-49F2-8690-3DAFCAE6FFB8}": r"%PROGRAMDATA%\Microsoft\Windows\Start Menu\Programs",
    "{9E3995AB-1F9C-4F13-B827-48B24B6C7174}": r"%AppData%\Roaming\Microsoft\Internet Explorer\Quick Launch\User Pinned",
    "{A77F5D77-2E2B-44C3-A6A2-ABA601054A51}": r"%AppData%\Roaming\Microsoft\Windows\Start Menu\Programs",
    "{D65231B0-B2F1-4857-A4CE-A8E7C6EA7D27}": r"%WINDIR%\SysWOW64",
}

WHITELISTED_NAMES = ["UEME_CTLSESSION"]

WIN_XP_USER_ASSIST = Struct(
    "session_id" / Int32ul,
    "run_counter" / Int32ul,
    "last_execution_timestamp" / Int64ul,
)

WIN7_USER_ASSIST = Struct(
    "session_id" / Int32ul,
    "run_counter" / Int32ul,
    "focus_count" / Int32ul,
    "total_focus_time_ms" / Int32ul,
    "unknown" * Bytes(44),
    "last_execution_timestamp" / Int64ul,
    "Const" * Const(b"\x00\x00\x00\x00"),
)


class UserAssistPlugin(Plugin):
    NAME = "user_assist"
    DESCRIPTION = "Parse User Assist artifact"
    COMPATIBLE_HIVE = NTUSER_HIVE_TYPE

    def run(self):
        for guid in GUIDS:
            try:
                subkey = self.registry_hive.get_key(
                    r"{}\{}".format(USER_ASSIST_KEY_PATH, guid)
                )
                count_subkey = subkey.get_subkey("Count")

                if not count_subkey.values_count:
                    logger.debug("Skipping {}".format(guid))
                    continue

                for value in count_subkey.iter_values(trim_values=False):
                    name = codecs.decode(value.name, encoding="rot-13")

                    if name in WHITELISTED_NAMES:
                        continue

                    for k, v in GUID_TO_PATH_MAPPINGS.items():
                        if k in name:
                            name = name.replace(k, v)
                            break

                    entry = None
                    data = value.value
                    if len(data) == 72:
                        try:
                            parsed_entry = WIN7_USER_ASSIST.parse(data)
                        except ConstError as ex:
                            logger.error(
                                f"Could not parse user assist entry named {name}: {ex}"
                            )
                            continue

                        entry = {
                            "name": name,
                            "timestamp": convert_wintime(
                                parsed_entry.last_execution_timestamp,
                                as_json=self.as_json,
                            ),
                            "run_counter": parsed_entry.run_counter,
                            "focus_count": parsed_entry.focus_count,
                            "total_focus_time_ms": parsed_entry.total_focus_time_ms,
                            "session_id": parsed_entry.session_id,
                        }

                    elif len(data) == 16:
                        try:
                            parsed_entry = WIN_XP_USER_ASSIST.parse(data)
                        except ConstError as ex:
                            logger.error(
                                f"Could not parse user assist entry named {name}: {ex}"
                            )
                            continue

                        entry = {
                            "name": name,
                            "timestamp": convert_wintime(
                                parsed_entry.last_execution_timestamp,
                                as_json=self.as_json,
                            ),
                            "session_id": parsed_entry.session_id,
                            "run_counter": parsed_entry.run_counter - 5,
                        }

                    if entry:
                        self.entries.append(entry)
            except RegistryKeyNotFoundException:
                continue



import logging

import attr
from regipy.exceptions import (
    RegistryKeyNotFoundException,
    RegistryParsingException,
)
from regipy.hive_types import SYSTEM_HIVE_TYPE
from regipy.plugins.plugin import Plugin
from regipy.utils import convert_wintime

logger = logging.getLogger(__name__)

SERVICES_PATH = r"Services"


class ServicesPlugin(Plugin):
    NAME = "services"
    DESCRIPTION = "Enumerate the services in the SYSTEM hive"
    COMPATIBLE_HIVE = SYSTEM_HIVE_TYPE

    def run(self):
        self.entries = {}
        logger.debug("Started Services enumeration Plugin...")
        for control_set_services_path in self.registry_hive.get_control_sets(
            SERVICES_PATH
        ):
            try:
                subkey = self.registry_hive.get_key(control_set_services_path)
            except RegistryKeyNotFoundException as ex:
                logger.error(ex)
                continue
            self.entries[control_set_services_path] = {
                "timestamp": convert_wintime(
                    subkey.header.last_modified, as_json=self.as_json
                )
            }
            services = []
            for service in subkey.iter_subkeys():
                values = None
                parameters = []
                if service.values_count > 0:
                    try:
                        values = [
                            attr.asdict(x) for x in service.iter_values(as_json=True)
                        ]
                    except RegistryParsingException as ex:
                        logger.error(
                            f"Exception while parsing data for service {service.name[:10] if service.name else None}: {ex}"
                        )

                    if service.subkey_count:
                        try:
                            service_parameters_path = r"{}\{}".format(
                                control_set_services_path, service.name
                            )
                            for parameter in self.registry_hive.recurse_subkeys(
                                nk_record=service,
                                path_root=service_parameters_path,
                                as_json=True,
                            ):
                                parameters.append(attr.asdict(parameter))
                        except RegistryParsingException as ex:
                            logger.info(
                                f"Exception while parsing parameters for service {service.name}: {ex}"
                            )

                entry = {
                    "name": service.name,
                    "last_modified": convert_wintime(
                        service.header.last_modified, as_json=self.as_json
                    ),
                    "values": values,
                    "parameters": parameters,
                }
                services.append(entry)
            self.entries[control_set_services_path]["services"] = services



import json
from tempfile import mktemp

from click.testing import CliRunner

from regipy.cli import parse_header, registry_dump, run_plugins


def test_cli_registry_parse_header(ntuser_hive):
    runner = CliRunner()
    result = runner.invoke(parse_header, [ntuser_hive])
    assert result.exit_code == 0
    assert len(result.output.splitlines()) == 27


def test_cli_registry_dump(ntuser_hive):
    runner = CliRunner()

    start_date = "2012-04-03T00:00:00.000000"
    end_date = "2012-04-03T23:59:59.999999"

    output_file_path = mktemp()
    result = runner.invoke(
        registry_dump,
        [ntuser_hive, "-d", "-s", start_date, "-e", end_date, "-o", output_file_path],
    )
    assert result.exit_code == 0

    with open(output_file_path, "r") as f:
        output = f.readlines()

    assert json.loads(output[0]) == {
        "subkey_name": ".Default",
        "path": "\\AppEvents\\EventLabels\\.Default",
        "timestamp": "2012-04-03T21:19:54.733216+00:00",
        "values_count": 2,
        "values": [],
        "actual_path": None,
    }

    assert json.loads(output[-1]) == {
        "subkey_name": "System",
        "path": "\\System",
        "timestamp": "2012-04-03T21:19:54.847482+00:00",
        "values_count": 0,
        "values": [],
        "actual_path": None,
    }

    # The output file contains one extra line, because of line breaks
    assert len(output) == 1490
    assert result.output.strip().endswith("(1489 subkeys enumerated)")


def test_cli_run_plugins(ntuser_hive):
    runner = CliRunner()

    output_file_path = mktemp()
    result = runner.invoke(run_plugins, [ntuser_hive, "-o", output_file_path])
    assert result.exit_code == 0

    assert (
        result.output.strip()
        == "Loaded 52 plugins\nFinished: 12/52 plugins matched the hive type"
    )

    with open(output_file_path, "r") as f:
        output = json.loads(f.read())

    assert set(output.keys()) == {
        "word_wheel_query",
        "winrar_plugin",
        "user_assist",
        "ntuser_persistence",
        "ntuser_classes_installer",
        "ntuser_shellbag_plugin",
        "terminal_services_history",
        "typed_paths",
        "network_drives_plugin",
        "typed_urls",
        "winscp_saved_sessions",
        "installed_programs_ntuser"
    }



import logging

from regipy.exceptions import RegistryKeyNotFoundException
from regipy.hive_types import NTUSER_HIVE_TYPE
from regipy.plugins.plugin import Plugin
from regipy.utils import convert_wintime

logger = logging.getLogger(__name__)

WINRAR_ARCHIVE_CREATION_HIST = r"\SOFTWARE\WinRAR\DialogEditHistory\ArcName"
WINRAR_ARCHIVE_EXTRACT_HIST = r"\SOFTWARE\WinRAR\DialogEditHistory\ExtrPath"
WINRAR_ARCHIVE_OPEN_HIST = r"\SOFTWARE\WinRAR\ArcHistory"


class WinRARPlugin(Plugin):
    NAME = "winrar_plugin"
    DESCRIPTION = "Parse the WinRAR archive history"
    COMPATIBLE_HIVE = NTUSER_HIVE_TYPE

    def run(self):
        try:
            open_subkey = self.registry_hive.get_key(WINRAR_ARCHIVE_OPEN_HIST)

            timestamp = convert_wintime(
                open_subkey.header.last_modified, as_json=self.as_json
            )
            for value in open_subkey.iter_values(as_json=self.as_json):
                self.entries.append(
                    {
                        "last_write": timestamp,
                        "file_path": value.value,
                        "value_name": value.name,
                        "operation": "archive_opened",
                    }
                )

        except RegistryKeyNotFoundException as ex:
            logger.error(
                f"Could not find {self.NAME} plugin data at: {WINRAR_ARCHIVE_OPEN_HIST}: {ex}"
            )

        try:
            create_subkey = self.registry_hive.get_key(WINRAR_ARCHIVE_CREATION_HIST)

            timestamp = convert_wintime(
                create_subkey.header.last_modified, as_json=self.as_json
            )
            for value in create_subkey.iter_values(as_json=self.as_json):
                self.entries.append(
                    {
                        "last_write": timestamp,
                        "file_name": value.value,
                        "value_name": value.name,
                        "operation": "archive_created",
                    }
                )

        except RegistryKeyNotFoundException as ex:
            logger.error(
                f"Could not find {self.NAME} plugin data at: {WINRAR_ARCHIVE_CREATION_HIST}: {ex}"
            )

        try:
            extract_subkey = self.registry_hive.get_key(WINRAR_ARCHIVE_EXTRACT_HIST)

            timestamp = convert_wintime(
                extract_subkey.header.last_modified, as_json=self.as_json
            )
            for value in extract_subkey.iter_values(as_json=self.as_json):
                self.entries.append(
                    {
                        "last_write": timestamp,
                        "file_path": value.value,
                        "value_name": value.name,
                        "operation": "archive_extracted",
                    }
                )

        except RegistryKeyNotFoundException as ex:
            logger.error(
                f"Could not find {self.NAME} plugin data at: {WINRAR_ARCHIVE_EXTRACT_HIST}: {ex}"
            )



import logging

from regipy.hive_types import SOFTWARE_HIVE_TYPE
from regipy.plugins.plugin import Plugin
from regipy.utils import convert_wintime
from regipy.exceptions import RegistryKeyNotFoundException

logger = logging.getLogger(__name__)


SPP_CLIENT_PATH = r"\Microsoft\Windows NT\CurrentVersion\SPP\Clients"
value_list = "{09F7EDC5-294E-4180-AF6A-FB0E6A0E9513}"


class SppClientsPlugin(Plugin):
    NAME = "spp_clients_plugin"
    DESCRIPTION = "Determines volumes monitored by VSS"
    COMPATIBLE_HIVE = SOFTWARE_HIVE_TYPE

    def can_run(self):
        return self.registry_hive.hive_type == SOFTWARE_HIVE_TYPE

    def run(self):
        logger.info("Started spp_clients Plugin...")

        try:
            key = self.registry_hive.get_key(SPP_CLIENT_PATH)
        except RegistryKeyNotFoundException as ex:
            logger.error(
                f"Could not find {self.NAME} subkey at {SPP_CLIENT_PATH}: {ex}"
            )
            return None

        self.entries = {
            SPP_CLIENT_PATH: {
                "last_write": convert_wintime(key.header.last_modified).isoformat()
            }
        }

        for val in key.iter_values():
            if val.name in value_list:
                aux_list = []
                for value in val.value:
                    aux_list.append(value.replace("%3A", ":"))
                self.entries[SPP_CLIENT_PATH][val.name] = aux_list



NTUSER_HIVE_TYPE = "ntuser"
SYSTEM_HIVE_TYPE = "system"
AMCACHE_HIVE_TYPE = "amcache"
SOFTWARE_HIVE_TYPE = "software"
SAM_HIVE_TYPE = "sam"
SECURITY_HIVE_TYPE = "security"
CLASSES_ROOT_HIVE_TYPE = "classes_root"
BCD_HIVE_TYPE = "bcd"
USRCLASS_HIVE_TYPE = "usrclass"


SUPPORTED_HIVE_TYPES = [
    NTUSER_HIVE_TYPE,
    SYSTEM_HIVE_TYPE,
    AMCACHE_HIVE_TYPE,
    SOFTWARE_HIVE_TYPE,
    SAM_HIVE_TYPE,
    SECURITY_HIVE_TYPE,
    BCD_HIVE_TYPE,
    USRCLASS_HIVE_TYPE,
]


HVLE_TRANSACTION_LOG_MAGIC = b"HvLE"
DIRT_TRANSACTION_LOG_MAGIC = b"DIRT"



import logging
from datetime import datetime, timezone
import re

from regipy.hive_types import SYSTEM_HIVE_TYPE
from regipy.plugins.plugin import Plugin
from regipy.exceptions import RegistryKeyNotFoundException

logger = logging.getLogger(__name__)


WIN_VER_PATH = r"\Setup"
os_list = [
    "ProductName",
    "ReleaseID",
    "CSDVersion",
    "CurrentVersion",
    "CurrentBuild",
    "CurrentBuildNumber",
    "InstallationType",
    "EditionID",
    "ProductName",
    "ProductId",
    "BuildLab",
    "BuildLabEx",
    "CompositionEditionID",
    "RegisteredOrganization",
    "RegisteredOwner",
    "InstallDate",
]


class PreviousWinVersionPlugin(Plugin):
    NAME = "previous_winver_plugin"
    DESCRIPTION = "Get previous relevant OS information"
    COMPATIBLE_HIVE = SYSTEM_HIVE_TYPE

    def can_run(self):
        return self.registry_hive.hive_type == SYSTEM_HIVE_TYPE

    def run(self):
        logger.info("Started winver Plugin...")

        try:
            key = self.registry_hive.get_key(WIN_VER_PATH)
        except RegistryKeyNotFoundException as ex:
            logger.error(f"Could not find {self.NAME} subkey at {WIN_VER_PATH}: {ex}")
            return None

        for sk in key.iter_subkeys():
            if sk.name.startswith("Source OS"):
                old_date = re.search(
                    r"Updated on (\d{1,2})/(\d{1,2})/(\d{4}) (\d{2}):(\d{2}):(\d{2})",
                    sk.name,
                )
                dt = datetime(
                    int(old_date.group(3)),
                    int(old_date.group(1)),
                    int(old_date.group(2)),
                    int(old_date.group(4)),
                    int(old_date.group(5)),
                    int(old_date.group(6)),
                ).strftime("%Y-%m-%d %H:%M:%S")
                temp_dict = {"key": f"\\{key.name}\\{sk.name}"}
                temp_dict["update_date"] = dt
                for val in sk.iter_values():
                    if val.name in os_list:
                        if val.name == "InstallDate":
                            temp_dict[val.name] = datetime.fromtimestamp(
                                val.value, timezone.utc
                            ).strftime("%Y-%m-%d %H:%M:%S")
                        else:
                            temp_dict[val.name] = val.value
                self.entries.append(temp_dict)
